# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

import click

from lean.click import LeanCommand
from lean.container import container
from lean.models.errors import MoreInfoError


@click.command(cls=LeanCommand)
@click.option("--user-id", "-u", type=str, help="QuantConnect user id")
@click.option("--api-token", "-t", type=str, help="QuantConnect API token")
def login(user_id: Optional[str], api_token: Optional[str]) -> None:
    """Log in with a QuantConnect account.

    If user id or API token is not provided an interactive prompt will show.

    Credentials are stored in ~/.lean/credentials and are removed upon running `lean logout`.
    """
    logger = container.logger()
    credentials_storage = container.credentials_storage()

    if user_id is None or api_token is None:
        logger.info("Your user id and API token are needed to make authenticated requests to the QuantConnect API")
        logger.info("You can request these credentials on https://www.quantconnect.com/account")
        logger.info(f"Both will be saved in {credentials_storage.file}")

    if user_id is None:
        user_id = click.prompt("User id")

    if api_token is None:
        api_token = logger.prompt_password("API token")

    api_client = container.api_client(user_id=user_id, api_token=api_token)
    if not api_client.is_authenticated():
        raise MoreInfoError("Credentials are invalid. Please ensure your computer clock is correct, or try using another terminal, or enter API token manually instead of copy-pasting.",
                            "https://www.lean.io/docs/v2/lean-cli")

    cli_config_manager = container.cli_config_manager()
    cli_config_manager.user_id.set_value(user_id)
    cli_config_manager.api_token.set_value(api_token)

    logger.info("Successfully logged in")



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import json
import time
import requests
from pathlib import Path

json_modules = {}
file_name = "modules-1.6.json"
directory = Path(__file__).parent
file_path = directory.parent / file_name

# check if new file is available online
url = f"https://cdn.quantconnect.com/cli/{file_name}"
error = None
try:
    # fetch if file not available or fetched before 1 day
    if not os.path.exists(file_path) or (time.time() - os.path.getmtime(file_path) >  86400):
        res = requests.get(url, timeout=5)
        if res.ok:
            new_content = res.json()
            with open(file_path, 'w', encoding='utf-8') as f:
                json.dump(new_content, f, ensure_ascii=False, indent=4)
        else:
            res.raise_for_status()
except Exception as e:
    # No need to do anything if file isn't available
    error = str(e)
    pass

# check if file exists
if not Path(file_path).is_file():
    error_message = f": {error}" if error is not None else ""
    raise FileNotFoundError(
        f"Modules json not found in the given path {file_path}{error_message}")

with open(file_path) as f:
    data = json.load(f)
    json_modules = data['modules']



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List

from lean.components.api.api_client import *


class ModuleClient:
    """The ModuleClient class contains methods to interact with modules/* API endpoints."""

    def __init__(self, api_client: 'APIClient') -> None:
        """Creates a new ModuleClient instance.

        :param api_client: the APIClient instance to use when making requests
        """
        self._api = api_client

    def get_link(self, product_id: int, organization_id: str, file_name: str) -> str:
        """Returns the download link to a module's file.

        :param product_id: the product id of the module
        :param organization_id: the id of the organization holding a license for the module
        :param file_name: the name of the module's file to download
        :return: the download link to the file
        """
        data = self._api.post("modules/read", {
            "productId": product_id,
            "organizationId": organization_id,
            "fileName": file_name
        })

        return data["url"]

    def list_files(self, product_id: int, organization_id: str) -> List[str]:
        """Lists the most recent files of a module.

        :param product_id: the product id of the module
        :param organization_id: the id of the organization holding a license for the module
        :return: a list of file names of recent files belonging to the given module
        """
        data = self._api.post("modules/list", {
            "productId": product_id,
            "organizationId": organization_id
        })

        return data["files"]



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from shutil import move
from tempfile import NamedTemporaryFile
from lean.components.api.api_client import *
from lean.models.api import QCDataInformation
from typing import List, Callable


class DataClient:
    """The DataClient class contains methods to interact with data/* API endpoints."""

    _list_files_cache: Dict[str, List[str]] = {}

    def __init__(self, api_client: 'APIClient', http_client: 'HTTPClient') -> None:
        """Creates a new DataClient instance.

        :param api_client: the APIClient instance to use when making requests
        :param http_client: the HTTPClient instance to use when downloading files
        """
        self._api = api_client
        self._http_client = http_client

    def download_file(self, relative_file_path: str, organization_id: str,
                      local_filename: str, progress_callback: Callable[[float], None]) -> None:
        """Downloads the content of a downloadable data file.

        :param relative_file_path: the relative path of the data file
        :param organization_id: the id of the organization that should be billed
        :param local_filename: the final local path where the data file will be stored
        :param progress_callback: the download progress callback
        """
        data = self._api.post("data/read", {
            "format": "link",
            "filePath": relative_file_path,
            "organizationId": organization_id
        })

        # we stream the data into a temporary file and later move it to it's final location
        with self._http_client.get(data["link"], stream=True) as r:
            r.raise_for_status()
            total_size = 0
            try:
                total_size = int(r.headers['Content-length'])
            except:
                pass
            current_size = 0
            with NamedTemporaryFile(delete=False) as f:
                temp_file_name = f.name
                for chunk in r.iter_content(chunk_size=1024 * 1024):
                    if total_size != 0:
                        previous = current_size / total_size

                    current_size += f.write(chunk)

                    if total_size != 0:
                        # progressive progress update if we can
                        progress_callback((current_size / total_size) - previous)
            if total_size == 0:
                # if total size not available update progress at the end
                progress_callback(1)

            directory = os.path.dirname(local_filename)
            os.makedirs(directory, exist_ok=True)
            move(temp_file_name, local_filename)

    def download_public_file(self, data_endpoint: str) -> bytes:
        """Downloads the content of a downloadable public file.

        :param data_endpoint: the url of the public file
        :return: the content of the file
        """
        return self._http_client.get(data_endpoint).content

    def list_files(self, prefix: str) -> List[str]:
        """Lists all remote files with a given prefix.

        :param prefix: the prefix of the files to return
        :return: the list of files with the given prefix
        """
        if prefix in DataClient._list_files_cache:
            return DataClient._list_files_cache[prefix]

        data = self._api.post("data/list", {
            "filePath": prefix
        })

        first_part = prefix.split("/")[0]
        files = sorted(f"{first_part}/{obj}" for obj in data["objects"])

        DataClient._list_files_cache[prefix] = files

        return files

    def get_info(self, organization_id: str) -> QCDataInformation:
        """Returns the available data vendors, their prices and a link to the data agreement.

        :param organization_id: the id of the organization to retrieve the data information for
        :return: the latest information on the downloadable data
        """
        data = self._api.post("data/prices", {
            "organizationId": organization_id
        })

        return QCDataInformation(**data)



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import click
from decimal import Decimal, DecimalException

class DecimalParamType(click.ParamType):
    name = "decimal"

    def convert(self, value, param, ctx):
        try:
            return Decimal(value)
        except DecimalException:
            self.fail(f"{value!r} is not a valid decimal", param, ctx)

DECIMAL = DecimalParamType()


# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random
from pathlib import Path
from typing import List

from lean.components.config.storage import Storage
from lean.components.util.xml_manager import XMLManager
from lean.components.util.path_manager import PathManager
from lean.constants import PROJECT_CONFIG_FILE_NAME
from lean.models.utils import CSharpLibrary


class ProjectConfigManager:
    """The ProjectConfigManager class manages the configuration of a single project."""

    def __init__(self, xml_manager: XMLManager) -> None:
        """Creates a new ProjectConfigManager instance.

        :param xml_manager: the XMLManager instance to use when parsing XML files
        """
        self._xml_manager = xml_manager

    def try_get_project_config(self, project_directory: Path, path_manager: PathManager) -> Storage:
        """Returns a Storage instance to get/set the configuration for a project.

        :param project_directory: the path to the project to retrieve the configuration of
        :return: the Storage instance containing the project-specific configuration of the given project
        """
        if path_manager.is_path_valid(project_directory) \
                and project_directory.is_dir() \
                and self.get_project_config(project_directory).file.exists():
            return Storage(str(project_directory / PROJECT_CONFIG_FILE_NAME))
        else:
            return False

    def get_project_config(self, project_directory: Path) -> Storage:
        """Returns a Storage instance to get/set the configuration for a project.

        :param project_directory: the path to the project to retrieve the configuration of
        :return: the Storage instance containing the project-specific configuration of the given project
        """
        return Storage(str(project_directory / PROJECT_CONFIG_FILE_NAME))

    def get_local_id(self, project_directory: Path) -> int:
        """Returns the local id of a project.

        Every Lean CLI project has a unique local id, regardless of whether the project is synchronized with the cloud.

        :param project_directory: the path to the project to retrieve the local id of
        :return: the local id of the given project
        """
        project_config = self.get_project_config(project_directory)

        if project_config.has("local-id"):
            return project_config.get("local-id")

        project_id = random.randint(100_000_000, 999_999_999)
        project_config.set("local-id", project_id)

        return project_id

    def get_latest_live_directory(self, project_directory: Path) -> Path:
        """Returns the path of the latest live directory.

        :param project_directory: the path to the project to retrieve the local id of
        :return: the path of the latest live directory
        """
        live_root_dir = project_directory / "live"
        if not live_root_dir.is_dir():
            raise ValueError(f"live directory {live_root_dir} is not a directory")

        for live_dir in sorted(list(live_root_dir.iterdir()), reverse=True):
            if not (live_dir / "log.txt").is_file():
                continue
            return live_dir
        
        raise ValueError("live directory with log.txt not found")

        
    def get_csharp_libraries(self, project_directory: Path) -> List[CSharpLibrary]:
        """Returns the custom C# libraries in a project.

        :param project_directory: the path to the project to retrieve the C# libraries of
        :return: a list containing the information of all PackageReferences in the project's .csproj file, if any
        """
        csproj_file = next((p for p in project_directory.iterdir() if p.name.endswith(".csproj")), None)
        if csproj_file is None:
            return []

        libraries = []
        csproj_tree = self._xml_manager.parse(csproj_file.read_text(encoding="utf-8"))

        for package_reference in csproj_tree.findall(".//PackageReference"):
            name = package_reference.get("Include", None)
            version = package_reference.get("Version", None)

            if name is not None and version is not None:
                libraries.append(CSharpLibrary(name=name, version=version))

        return libraries



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.



import click
from lean.click import LeanCommand
from lean.container import container


@click.command(cls=LeanCommand)
@click.argument("project", type=str)
def stop(project: str) -> None:
    """
    Stops live trading for a certain project without liquidating existing positions.
    """
    logger = container.logger()
    api_client = container.api_client()

    cloud_project_manager = container.cloud_project_manager()
    cloud_project = cloud_project_manager.get_cloud_project(project, False)
    logger.info(f"cloud.live.stop(): sending command.")
    response = api_client.live.stop(cloud_project.projectId)
    if response.success:
        logger.info(f"cloud.live.stop(): command executed successfully.")
    else:
        raise Exception("cloud.live.stop(): Failed: to execute the command successfully.")





# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

from lean.components.api.api_client import *
from lean.models.api import QCUser


class UserClient:
    """The UserClient class contains methods to interact with user/* API endpoints."""

    def __init__(self, api_client: 'APIClient') -> None:
        """Creates a new UserClient instance.

        :param api_client: the APIClient instance to use when making requests
        """
        self._api = api_client

    def get_info(self, organization_id: Optional[str] = None) -> QCUser:
        """Returns the user's name, avatar url and organization badge.

        :param organization_id: the id of the organization to get the badge of or None to use the preferred one
        :return: the user's name, avatar url and organization badge
        """
        parameters = {}

        if organization_id is not None:
            parameters["organizationId"] = organization_id

        data = self._api.post("user/read", parameters)
        return QCUser(**data["user"])



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
from typing import Any, Dict
import click
from lean.components.util.click_group_default_command import DefaultCommandGroup
from lean.constants import COMMAND_FILE_BASENAME, COMMAND_RESULT_FILE_BASENAME
import time
from pathlib import Path
from lean.container import container


@click.group(cls=DefaultCommandGroup)
def live() -> None:
    """Interact with the local machine."""
    # This method is intentionally empty
    # It is used as the command group for all `lean cloud <command>` commands
    pass


def get_command_file_name() -> str:
    return Path(f'{COMMAND_FILE_BASENAME}-{int(time.time())}.json')


def get_result_file_name(command_id: str) -> str:
    return Path(f'{COMMAND_RESULT_FILE_BASENAME}-{command_id}.json')


def send_command(project: Path, data: Dict[str, Any]) -> str:
    """send command to the running container of the given project.

    :param project: the project path
    :param data: the data to send to the container
    :return: the name of the running docker container of the given project
    """
    logger = container.logger()
    live_dir = container.project_config_manager().get_latest_live_directory(project)
    docker_container_name = container.output_config_manager(
    ).get_container_name(Path(live_dir))
    file_name = get_command_file_name()
    logger.info(
        f"live.send_command(): {inspect.stack()[1].function} - sending command.")
    container.docker_manager().write_to_file(
        docker_container_name, file_name, data)
    return docker_container_name


def get_result(command_id: str, docker_container_name: str, container_running_required: bool = True,
               interval: int = 1, timeout: int = 30) -> None:
    """Get the result of a command.

    :param command_id: command id
    :param docker_container_name: docker container name
    :param container_running_required: should the container be alive after command execution, defaults to True
    :param interval: interval to sleep before retrying, defaults to 1
    :param timeout: time to stop trying to check for result file, defaults to 30
    :raises Exception: When the command is not executed successfully
    """
    logger = container.logger()
    logger.info(
        f"live.get_result(): {inspect.stack()[1].function} -  waiting for results...")
    result_file_path = get_result_file_name(command_id)
    result = container.docker_manager().read_from_file(
        docker_container_name, result_file_path, interval, timeout)
    if "success" in result and result["success"]:
        logger.info(
            f"live.get_result(): {inspect.stack()[1].function} - Success: The command was executed successfully")
    elif "container-running" in result and not result["container-running"]:
        if container_running_required:
            raise Exception(
                f"live.get_result(): {inspect.stack()[1].function} - Failed: The container is not running")
        else:
            logger.info(
                f"live.get_result(): {inspect.stack()[1].function} - Success: The command was executed successfully")
    else:
        raise Exception((f"live.get_result(): {inspect.stack()[1].function} - "
                        + f"Failed: to execute the command successfully. {result['error']}"))



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Optional

from lean.components.config.storage import Storage


class Option:
    """An Option instance manages a single value in a Storage instance."""

    def __init__(self, key: str, description: str, is_sensitive: bool, storage: Storage) -> None:
        """Creates a new StringOption instance.

        :param key: the name of the key of the option in the given Storage instance, using hyphens for separation
        :param description: a display-friendly description of the option
        :param is_sensitive: whether the contents of this option may be logged without masking it
        :param storage: the Storage instance to store this option in
        """
        self.key = key
        self.description = description
        self.is_sensitive = is_sensitive

        self._storage = storage
        self.location = storage.file

    def get_value(self, default: Optional[str] = None) -> Optional[str]:
        """Retrieves the current value of the option.

        :param default: the value to return if the option is not set
        :return: the current value of the option, or the given default if the option is not set
        """
        return self._storage.get(self.key, default)

    def set_value(self, value: str) -> None:
        """Sets the new value of the option.

        :param value: the non-empty new value of the option
        """
        if value == "":
            raise ValueError("Value cannot be empty")

        self._storage.set(self.key, value)

    def unset(self) -> None:
        """Unsets any value of the option."""
        self._storage.delete(self.key)


class ChoiceOption(Option):
    """A variant of Option where only a few values are allowed.

    The allowed values are considered to be case insensitive.
    """

    def __init__(self,
                 key: str,
                 description: str,
                 allowed_values: List[str],
                 is_sensitive: bool,
                 storage: Storage,
                 default_value: str = None) -> None:
        """Creates a new ChoiceOption instance.

        :param key: the name of the key of the option in the given file, should use hyphens for separation
        :param description: a display-friendly description of the option
        :param allowed_values: the values which can be set
        :param is_sensitive: whether the contents of this option may be logged without masking it
        :param storage: the Storage instance to store this option in
        :param default_value: the default value to use for the choices
        """
        self.allowed_values = allowed_values
        self.default_value = default_value

        if description.endswith("."):
            description = description[:-1]
        description = description + f" (allowed values: {', '.join(allowed_values)})."

        super().__init__(key, description, is_sensitive, storage)

    def set_value(self, value: str) -> None:
        """Sets the new value of the option.

        :param value: the new value of the option, must be one of this option's allowed values
        """
        matching_value = next((x for x in self.allowed_values if x.lower() == value.lower()), None)

        if matching_value is None:
            raise ValueError(
                f"Invalid value, '{self.key}' only accepts the following values: {', '.join(self.allowed_values)}")

        super().set_value(matching_value)



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from lean.commands.live.live import live
from lean.commands.live.deploy import deploy
from lean.commands.live.stop import stop
from lean.commands.live.liquidate import liquidate
from lean.commands.live.submit_order import submit_order
from lean.commands.live.cancel_order import cancel_order
from lean.commands.live.add_security import add_security
from lean.commands.live.update_order import update_order

live.add_command(deploy)
live.add_command(stop)
live.add_command(liquidate)
live.add_command(submit_order)
live.add_command(cancel_order)
live.add_command(add_security)
live.add_command(update_order)


# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import re
import urllib.request

from setuptools import find_packages, setup

current_dir = os.path.abspath(os.path.dirname(__file__))


def read(relative_path: str) -> str:
    with open(os.path.join(current_dir, relative_path)) as file:
        return file.read()


def get_version() -> str:
    version_file = read("lean/__init__.py")
    version_match = re.search(r"^__version__ = \"([^\"]+)\"", version_file, re.M)
    return version_match.group(1)


def get_stubs_version_range() -> str:
    if get_version() == "dev":
        return ""

    try:
        response = urllib.request.urlopen("https://pypi.org/pypi/quantconnect-stubs/json").read()
        latest_version = json.loads(response)["info"]["version"]
        return f">={latest_version}"
    except:
        return ""


# Production dependencies
install_requires = [
    "click>=8.0.4",
    "requests>=2.27.1",
    "json5>=0.9.8",
    "docker>=6.0.0",
    "rich>=9.10.0",
    "dependency-injector>=4.39.1",
    "pydantic>=1.8.2",
    "python-dateutil>=2.8.2",
    "lxml>=4.9.0",
    "maskpass==0.3.6",
    "joblib>=1.1.0",
    "wrapt~=1.14.1",
    "setuptools",
    f"quantconnect-stubs{get_stubs_version_range()}"
]

setup(
    name="lean",
    version=get_version(),
    description="A CLI aimed at making it easier to run QuantConnect's LEAN engine locally and in the cloud",
    author="QuantConnect",
    author_email="support@quantconnect.com",
    url="https://lean.io/cli",
    long_description=read("README.md").replace("](lean", "](https://github.com/QuantConnect/lean-cli/blob/master/lean"),
    long_description_content_type="text/markdown",
    packages=find_packages(include=["lean", "lean.*"]),
    package_data={
        "lean": ["ssh/*"]
    },
    entry_points={
        "console_scripts": ["lean=lean.main:main"]
    },
    install_requires=install_requires,
    python_requires=">= 3.7",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Intended Audience :: Financial and Insurance Industry",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10"
    ],
    project_urls={
        "Documentation": "https://www.lean.io/docs/v2/lean-cli/key-concepts/getting-started",
        "Source": "https://github.com/QuantConnect/lean-cli"
    },
)



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from datetime import datetime
from enum import Enum
from typing import Any, Dict, List, Optional, Union

from pydantic import validator
from rich import box
from rich.table import Table
from rich.text import Text

from lean.constants import EQUITY_SECURITY_MASTER_PRODUCT_ID, BULK_EQUITY_SECURITY_MASTER_PRODUCT_ID
from lean.models.pydantic import WrappedBaseModel


# The models in this module are all parts of responses from the QuantConnect API
# The keys of properties are not changed, so they don't obey the rest of the project's naming conventions


class QCCollaborator(WrappedBaseModel):
    id: int
    uid: int
    blivecontrol: bool
    epermission: str
    profileimage: str
    name: str
    owner: bool = False


class QCParameter(WrappedBaseModel):
    key: str
    value: str
    min: Optional[float]
    max: Optional[float]
    step: Optional[float]
    type: Optional[str]


class QCLiveResults(WrappedBaseModel):
    eStatus: str
    sDeployID: Optional[str] = None
    sServerType: Optional[str] = None
    dtLaunched: Optional[datetime] = None
    dtStopped: Optional[datetime] = None
    sBrokerage: Optional[str] = None
    sSecurityTypes: Optional[str] = None
    dUnrealized: Optional[float] = None
    dfees: Optional[float] = None
    dnetprofit: Optional[float] = None
    dEquity: Optional[float] = None
    dHoldings: Optional[float] = None
    dCapital: Optional[float] = None
    dVolume: Optional[float] = None
    iTrades: Optional[int] = None
    sErrorMessage: Optional[str] = None


class QCLanguage(str, Enum):
    CSharp = "C#"
    FSharp = "F#"
    VisualBasic = "VB"
    Java = "Ja"
    Python = "Py"


class QCProject(WrappedBaseModel):
    projectId: int
    organizationId: str
    name: str
    description: str
    modified: datetime
    created: datetime
    language: QCLanguage
    collaborators: List[QCCollaborator]
    leanVersionId: int
    leanPinnedToMaster: bool
    leanEnvironment: int
    parameters: List[QCParameter]
    liveResults: QCLiveResults
    libraries: List[int]

    @validator("parameters", pre=True)
    def process_parameters_dict(cls, value: Any) -> Any:
        if isinstance(value, dict):
            return list(value.values())
        return value

    def get_url(self) -> str:
        """Returns the url of the project page in the cloud.

        :return: a url which when visited opens an Algorithm Lab tab containing the project
        """
        return f"https://www.quantconnect.com/project/{self.projectId}"

    def __hash__(self):
        return hash(self.projectId)

    def __eq__(self, other: Any):
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.projectId == other.projectId


class QCCreatedProject(WrappedBaseModel):
    projectId: int
    name: str
    modified: datetime
    created: datetime


class QCFullFile(WrappedBaseModel):
    name: str
    content: str
    modified: datetime
    isLibrary: bool


class QCMinimalFile(WrappedBaseModel):
    name: str
    content: str
    modified: datetime


class QCCompileParameter(WrappedBaseModel):
    line: int
    type: str


class QCCompileParameterContainer(WrappedBaseModel):
    file: str
    parameters: List[QCCompileParameter]


class QCCompileState(str, Enum):
    InQueue = "InQueue"
    BuildSuccess = "BuildSuccess"
    BuildError = "BuildError"


class QCCompileWithLogs(WrappedBaseModel):
    compileId: str
    state: QCCompileState
    logs: List[str]


class QCCompileWithParameters(WrappedBaseModel):
    compileId: str
    state: QCCompileState
    parameters: List[QCCompileParameterContainer]


class QCBacktest(WrappedBaseModel):
    backtestId: str
    projectId: int
    status: str
    name: str
    note: Optional[str] = None
    created: datetime
    completed: bool
    progress: float
    result: Optional[Any] = None
    error: Optional[str] = None
    stacktrace: Optional[str] = None
    runtimeStatistics: Optional[Dict[str, str]]
    statistics: Optional[Union[Dict[str, str], List[Any]]]
    totalPerformance: Optional[Any]

    def is_complete(self) -> bool:
        """Returns whether the backtest has completed in the cloud.

        :return: True if the backtest is complete, False if not
        """
        if self.error is not None:
            return True

        if not self.completed:
            return False

        has_runtime_statistics = self.runtimeStatistics is not None
        has_statistics = self.statistics is not None and not isinstance(self.statistics, list)
        return has_runtime_statistics and has_statistics

    def get_url(self) -> str:
        """Returns the url of the backtests results page in the cloud.

        :return: a url which when visited opens an Algorithm Lab tab containing the backtest's results
        """
        return f"https://www.quantconnect.com/project/{self.projectId}/{self.backtestId}"

    def get_statistics_table(self) -> Table:
        """Converts the statistics into a pretty table.

        :return: a table containing all statistics
        """
        stats = []

        for key, value in self.runtimeStatistics.items():
            stats.append(key)

            if "-" in value:
                stats.append(Text.from_markup(f"[red]{value}[/red]"))
            elif any(char.isdigit() and int(char) > 0 for char in value):
                stats.append(Text.from_markup(f"[green]{value}[/green]"))
            else:
                stats.append(value)

        if len(stats) % 4 != 0:
            stats.extend(["", ""])

        end_of_first_section = len(stats)

        for key, value in self.statistics.items():
            stats.extend([key, value])

        if len(stats) % 4 != 0:
            stats.extend(["", ""])

        table = Table(box=box.SQUARE)
        table.add_column("Statistic", overflow="fold")
        table.add_column("Value", overflow="fold")
        table.add_column("Statistic", overflow="fold")
        table.add_column("Value", overflow="fold")

        for i in range(int(len(stats) / 4)):
            start = i * 4
            end = (i + 1) * 4
            table.add_row(*stats[start:end], end_section=end_of_first_section == end)

        return table


class QCBacktestReport(WrappedBaseModel):
    report: str


class QCNodePrice(WrappedBaseModel):
    monthly: int
    yearly: int


class QCNode(WrappedBaseModel):
    id: str
    name: str
    projectName: str
    description: str
    usedBy: str
    sku: str
    busy: bool
    price: QCNodePrice
    speed: float
    cpu: int
    ram: float
    assets: int
    host: Optional[str]


class QCNodeList(WrappedBaseModel):
    backtest: List[QCNode]
    research: List[QCNode]
    live: List[QCNode]


class QCLiveAlgorithmStatus(str, Enum):
    DeployError = "DeployError"
    InQueue = "InQueue"
    Running = "Running"
    Stopped = "Stopped"
    Liquidated = "Liquidated"
    Deleted = "Deleted"
    Completed = "Completed"
    RuntimeError = "RuntimeError"
    Invalid = "Invalid"
    LoggingIn = "LoggingIn"
    Initializing = "Initializing"
    History = "History"


class QCRestResponse(WrappedBaseModel):
    success: bool
    error: Optional[List[str]]

class QCMinimalLiveAlgorithm(WrappedBaseModel):
    projectId: int
    deployId: str
    status: Optional[QCLiveAlgorithmStatus] = None

    def get_url(self) -> str:
        """Returns the url of the live deployment in the cloud.

        :return: a url which when visited opens an Algorithm Lab tab containing the live deployment
        """
        return f"https://www.quantconnect.com/project/{self.projectId}/live"


class QCFullLiveAlgorithm(QCMinimalLiveAlgorithm):
    projectId: int
    deployId: str
    status: Optional[QCLiveAlgorithmStatus] = None
    launched: datetime
    stopped: Optional[datetime]
    brokerage: str
    subscription: str
    error: str


class QCEmailNotificationMethod(WrappedBaseModel):
    address: str
    subject: str


class QCWebhookNotificationMethod(WrappedBaseModel):
    address: str
    headers: Dict[str, str]


class QCSMSNotificationMethod(WrappedBaseModel):
    phoneNumber: str


class QCTelegramNotificationMethod(WrappedBaseModel):
    id: str
    token: Optional[str] = None


QCNotificationMethod = Union[QCEmailNotificationMethod, QCWebhookNotificationMethod, QCSMSNotificationMethod, QCTelegramNotificationMethod]


class QCCard(WrappedBaseModel):
    brand: str
    expiration: str
    last4: str


class QCAccount(WrappedBaseModel):
    organizationId: str
    card: Optional[QCCard] = None

    # Balance in QCC
    creditBalance: float


class QCOrganizationCreditMovement(WrappedBaseModel):
    date: str
    description: str
    type: str
    subtype: str
    amount: float

    # Balance in QCC
    balance: float


class QCOrganizationCredit(WrappedBaseModel):
    movements: List[QCOrganizationCreditMovement]

    # Balance in QCC
    balance: float


class QCOrganizationProductItem(WrappedBaseModel):
    productId: int
    name: str
    quantity: int
    unitPrice: float
    total: float


class QCOrganizationProduct(WrappedBaseModel):
    name: str
    items: List[QCOrganizationProductItem]


class QCOrganizationData(WrappedBaseModel):
    signedTime: Optional[int]
    current: bool


class QCOrganizationMember(WrappedBaseModel):
    id: int
    name: str
    isAdmin: bool
    email: str


class QCFullOrganization(WrappedBaseModel):
    id: str
    name: str
    seats: int
    type: str
    credit: QCOrganizationCredit
    products: List[QCOrganizationProduct]
    data: QCOrganizationData
    members: List[QCOrganizationMember]

    def has_security_master_subscription(self) -> bool:
        """Returns whether this organization has a Security Master subscription.

        :return: True if the organization has a Security Master subscription, False if not
        """

        # TODO: This sort of hardcoded product ID checking is not sufficient when we consider
        # multiple 'Security Master' products. Especially since they will be specific to certain datasources.
        # For now, simple checks for an equity "Security Master" subscription
        # Issue created here: https://github.com/QuantConnect/lean-cli/issues/73

        data_products_product = next((x for x in self.products if x.name == "Data"), None)
        if data_products_product is None:
            return False

        return any(x.productId in {EQUITY_SECURITY_MASTER_PRODUCT_ID, BULK_EQUITY_SECURITY_MASTER_PRODUCT_ID} for x in data_products_product.items)


class QCMinimalOrganization(WrappedBaseModel):
    id: str
    name: str
    type: str
    ownerName: str
    members: int
    preferred: bool


class QCSecurityType(str, Enum):
    Equity = "Equity"
    Forex = "Forex"
    CFD = "Cfd"
    Future = "Future"
    Crypto = "Crypto"
    Option = "Option"
    FutureOption = "FutureOption"


class QCResolution(str, Enum):
    Tick = "Tick"
    Second = "Second"
    Minute = "Minute"
    Hour = "Hour"
    Daily = "Daily"

    @classmethod
    def by_name(cls, name: str) -> 'QCResolution':
        """Returns the enum member with the same name as the given one, case insensitively.

        :param name: the name of the enum member (case insensitive)
        :return: the matching enum member
        """
        for k, v in cls.__members__.items():
            if k.lower() == name.lower():
                return v
        raise ValueError(f"QCResolution has no member named '{name}'")


class QCLink(WrappedBaseModel):
    link: str


class QCOptimizationBacktest(WrappedBaseModel):
    id: str
    name: str
    exitCode: int
    parameterSet: Dict[str, str]
    statistics: List[float] = []


class QCOptimization(WrappedBaseModel):
    optimizationId: str
    projectId: int
    status: str
    name: str
    backtests: Dict[str, QCOptimizationBacktest] = {}
    runtimeStatistics: Dict[str, str] = {}

    @validator("backtests", "runtimeStatistics", pre=True)
    def parse_empty_lists(cls, value: Any) -> Any:
        # If these fields have no data, they are assigned an array by default
        # For consistency we convert those empty arrays to empty dicts
        if isinstance(value, list):
            return {}
        return value

    def get_progress(self) -> float:
        """Returns the progress of the optimization between 0.0 and 1.0.

        :return: 0.0 if the optimization is 0% done, 1.0 if the optimization is 100% done, or somewhere in between
        """
        stats = self.runtimeStatistics
        if "Completed" in stats and "Failed" in stats and "Total" in stats:
            finished_backtests = float(stats["Completed"]) + float(stats["Failed"])
            total_backtests = float(stats["Total"])
            return finished_backtests / total_backtests
        return 0.0


class QCOptimizationEstimate(WrappedBaseModel):
    estimateId: str
    time: int
    balance: int


class QCDataVendor(WrappedBaseModel):
    vendorName: str
    regex: Any

    # Price in QCC
    price: Optional[float]

    @validator("regex", pre=True)
    def parse_regex(cls, value: Any) -> Any:
        if isinstance(value, str):
            return re.compile(value[value.index("/") + 1:value.rindex("/")])
        return value


class QCDataInformation(WrappedBaseModel):
    datasources: Dict[str, Any]
    prices: List[QCDataVendor]
    agreement: str


class QCDatasetDelivery(str, Enum):
    CloudOnly = "cloud only"
    DownloadOnly = "download only"
    CloudAndDownload = "cloud & download"


class QCDatasetTag(WrappedBaseModel):
    name: str


class QCDataset(WrappedBaseModel):
    id: int
    name: str
    delivery: QCDatasetDelivery
    vendorName: str
    tags: List[QCDatasetTag]


class QCUser(WrappedBaseModel):
    name: str
    profile: str
    badge: Optional[str]


class QCTerminalNewsItem(WrappedBaseModel):
    id: int
    type: str
    category: str
    title: str
    content: str
    image: str
    link: str
    year_deleted: Optional[Any]
    week_deleted: Optional[Any]
    created: datetime
    date: datetime


class QCLeanEnvironment(WrappedBaseModel):
    id: int
    name: str
    path: Optional[str]
    description: str
    public: bool



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path
from typing import Optional, List

import click

from lean.click import LeanCommand, PathParameter
from lean.constants import PROJECT_CONFIG_FILE_NAME
from lean.container import container


def _get_project_directories() -> List[Path]:
    directories_to_check = [container.lean_config_manager().get_cli_root_directory()]
    project_directories = []

    while len(directories_to_check) > 0:
        directory = directories_to_check.pop(0)

        config_file = directory / PROJECT_CONFIG_FILE_NAME
        if config_file.is_file():
            project_directories.append(directory)
        else:
            directories_to_check.extend(d for d in directory.iterdir() if d.is_dir())

    return project_directories


@click.command(cls=LeanCommand, requires_lean_config=True)
@click.option("--backtest", is_flag=True, default=False, help="Display the most recent backtest logs (default)")
@click.option("--live", is_flag=True, default=False, help="Display the most recent live logs")
@click.option("--optimization", is_flag=True, default=False, help="Display the most recent optimization logs")
@click.option("--project",
              type=PathParameter(exists=True, file_okay=False, dir_okay=True),
              help="The project to get the most recent logs from")
def logs(backtest: bool, live: bool, optimization: bool, project: Optional[Path]) -> None:
    """Display the most recent backtest/live/optimization logs."""
    if [backtest, live, optimization].count(True) > 1:
        raise RuntimeError("--backtest, --live and --optimization are mutually exclusive")

    if not backtest and not live and not optimization:
        backtest = True

    if backtest:
        mode = "backtest"
        mode_directory = "backtests"
    elif live:
        mode = "live"
        mode_directory = "live"
    elif optimization:
        mode = "optimization"
        mode_directory = "optimizations"

    if project is None:
        project_directories = _get_project_directories()
    else:
        project_directories = [project]

    most_recent_file = None
    most_recent_timestamp = None

    for project_directory in project_directories:
        target_directory = project_directory / mode_directory
        if not target_directory.is_dir():
            continue

        for session_directory in target_directory.iterdir():
            if not session_directory.is_dir():
                continue

            log_file = session_directory / "log.txt"
            if not log_file.is_file():
                continue

            log_file_timestamp = log_file.stat().st_mtime_ns
            if most_recent_file is None or log_file_timestamp > most_recent_timestamp:
                most_recent_file = log_file
                most_recent_timestamp = log_file_timestamp

    if most_recent_file is None:
        raise RuntimeError(f"No {mode} log file exists")

    print(most_recent_file.read_text(encoding="utf-8").strip())



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List

from lean.components.api.api_client import *
from lean.models.api import QCBacktest, QCBacktestReport


class BacktestClient:
    """The BacktestClient class contains methods to interact with backtests/* API endpoints."""

    def __init__(self, api_client: 'APIClient') -> None:
        """Creates a new BacktestClient instance.

        :param api_client: the APIClient instance to use when making requests
        """
        self._api = api_client

    def get(self, project_id: int, backtest_id: str) -> QCBacktest:
        """Returns the details of a backtest.

        :param project_id: the id of the project the backtest belongs to
        :param backtest_id: the id of the backtest to retrieve the details of
        :return: the details of the specified backtest
        """
        data = self._api.get("backtests/read", {
            "projectId": project_id,
            "backtestId": backtest_id
        })

        return QCBacktest(**data["backtest"])

    def get_all(self, project_id: int) -> List[QCBacktest]:
        """Returns all backtests in a project.

        :param project_id: the id of the project to retrieve the backtests of
        :return: the backtests in the specified project
        """
        data = self._api.get("backtests/read", {
            "projectId": project_id
        })

        return [QCBacktest(**backtest) for backtest in data["backtests"]]

    def create(self, project_id: int, compile_id: str, name: str) -> QCBacktest:
        """Creates a new backtest.

        :param project_id: the id of the project to create a backtest for
        :param compile_id: the id of a compilation of the given project
        :param name: the name of the new backtest
        :return: the created backtest
        """
        data = self._api.post("backtests/create", {
            "projectId": project_id,
            "compileId": compile_id,
            "backtestName": name,
            "requestSource": f"CLI {lean.__version__}"
        })

        return QCBacktest(**data["backtest"])

    def get_report(self, project_id: int, backtest_id: str) -> QCBacktestReport:
        """Returns the report of a backtest.

        :param project_id: the id of the project the backtest belongs to
        :param backtest_id: the id of the backtest to retrieve the report of
        :return: the report of the specified backtest
        """
        data = self._api.post("backtests/read/report", {
            "projectId": project_id,
            "backtestId": backtest_id
        })

        return QCBacktestReport(**data)

    def update(self, project_id: int, backtest_id: str, name: str, note: str) -> None:
        """Updates an existing backtest.

        :param project_id: the id of the project the backtest belongs to
        :param backtest_id: the id of the backtest to update
        :param name: the new name to assign to the backtest
        :param note: the new note to assign to the backtest
        """
        self._api.post("backtests/update", {
            "projectId": project_id,
            "backtestId": backtest_id,
            "name": name,
            "note": note
        })

    def delete(self, project_id: int, backtest_id: str) -> None:
        """Deletes an existing backtest.

        :param project_id: the id of the project the backtest belongs to
        :param backtest_id: the id of the backtest to delete
        """
        self._api.post("backtests/delete", {
            "projectId": project_id,
            "backtestId": backtest_id
        })



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
import os
import shutil
import sys
from datetime import datetime
from pathlib import Path
from typing import Optional, List

import click

from lean.constants import DEFAULT_LEAN_CONFIG_FILE_NAME
from lean.container import container
from lean.models.errors import MoreInfoError
from lean.models.logger import Option


class LeanCommand(click.Command):
    """A click.Command wrapper with some Lean CLI customization."""

    def __init__(self,
                 requires_lean_config: bool = False,
                 requires_docker: bool = False,
                 allow_unknown_options: bool = False,
                 *args,
                 **kwargs):
        """Creates a new LeanCommand instance.

        :param requires_lean_config: True if this command requires a Lean config, False if not
        :param requires_docker: True if this command uses Docker, False if not
        :param allow_unknown_options: True if unknown options are allowed, False if not
        :param args: the args that are passed on to the click.Command constructor
        :param kwargs: the kwargs that are passed on to the click.Command constructor
        """
        self._requires_lean_config = requires_lean_config
        self._requires_docker = requires_docker
        self._allow_unknown_options = allow_unknown_options

        super().__init__(*args, **kwargs)

        # By default the width of help messages is min(terminal_width, max_content_width)
        # max_content_width defaults to 80, which we increase to 120 to improve readability on wide terminals
        self.context_settings["max_content_width"] = 120

        # Don't fail if unknown options are passed in when they're allowed
        self.context_settings["ignore_unknown_options"] = allow_unknown_options
        self.context_settings["allow_extra_args"] = allow_unknown_options

    def invoke(self, ctx: click.Context):
        if self._requires_lean_config:
            lean_config_manager = container.lean_config_manager()
            try:
                # This method will raise an error if the directory cannot be found
                lean_config_manager.get_cli_root_directory()
            except Exception:
                # Use one of the cached Lean config locations to avoid having to abort the command
                lean_config_paths = lean_config_manager.get_known_lean_config_paths()
                if len(lean_config_paths) > 0:
                    lean_config_path = container.logger().prompt_list("Select the Lean configuration file to use", [
                        Option(id=p, label=str(p)) for p in lean_config_paths
                    ])
                    lean_config_manager.set_default_lean_config_path(lean_config_path)
                else:
                    # Abort with a display-friendly error message if the command requires a Lean config and none found
                    raise MoreInfoError(
                        "This command requires a Lean configuration file, run `lean init` in an empty directory to create one, or specify the file to use with --lean-config",
                        "https://www.lean.io/docs/v2/lean-cli/key-concepts/troubleshooting#02-Common-Errors"
                    )

        if self._requires_docker and "pytest" not in sys.modules:
            is_system_linux = container.platform_manager().is_system_linux()

            # The CLI uses temporary directories in /tmp because sometimes it may leave behind files owned by root
            # These files cannot be deleted by the CLI itself, so we rely on the OS to empty /tmp on reboot
            # The Snap version of Docker does not provide access to files outside $HOME, so we can't support it
            if is_system_linux:
                docker_path = shutil.which("docker")
                if docker_path is not None and docker_path.startswith("/snap"):
                    raise MoreInfoError(
                        "The Lean CLI does not work with the Snap version of Docker, please re-install Docker via the official installation instructions",
                        "https://docs.docker.com/engine/install/")

            # A usual Docker installation on Linux requires the user to use sudo to run Docker
            # If we detect that this is the case and the CLI was started without sudo we elevate automatically
            if is_system_linux and os.getuid() != 0 and container.docker_manager().is_missing_permission():
                container.logger().info(
                    "This command requires access to Docker, you may be asked to enter your password")

                args = ["sudo", "--preserve-env=HOME", sys.executable, *sys.argv]
                os.execlp(args[0], *args)

        if self._allow_unknown_options:
            # Unknown options are passed to ctx.args and need to be parsed manually
            # We parse them to ctx.params so they're available like normal options
            # Because of this all commands with allow_unknown_options=True must have a **kwargs argument
            arguments = list(itertools.chain(*[arg.split("=") for arg in ctx.args]))

            skip_next = False
            for index in range(len(arguments) - 1):
                if skip_next:
                    skip_next = False
                    continue

                if arguments[index].startswith("--"):
                    option = arguments[index].replace("--", "")
                    value = arguments[index + 1]
                    ctx.params[option] = value
                    skip_next = True

        update_manager = container.update_manager()
        update_manager.show_announcements()

        result = super().invoke(ctx)

        update_manager.warn_if_cli_outdated()

        return result

    def get_params(self, ctx: click.Context):
        params = super().get_params(ctx)

        # Add --lean-config option if the command requires a Lean config
        if self._requires_lean_config:
            params.insert(len(params) - 1, click.Option(["--lean-config"],
                                                        type=PathParameter(exists=True, file_okay=True, dir_okay=False),
                                                        help=f"The Lean configuration file that should be used (defaults to the nearest {DEFAULT_LEAN_CONFIG_FILE_NAME})",
                                                        expose_value=False,
                                                        is_eager=True,
                                                        callback=self._parse_config_option))

        # Add --verbose option
        params.insert(len(params) - 1, click.Option(["--verbose"],
                                                    help="Enable debug logging",
                                                    is_flag=True,
                                                    default=False,
                                                    expose_value=False,
                                                    is_eager=True,
                                                    callback=self._parse_verbose_option))

        return params

    def _parse_config_option(self, ctx: click.Context, param: click.Parameter, value: Optional[Path]) -> None:
        """Parses the --config option."""
        if value is not None:
            lean_config_manager = container.lean_config_manager()
            lean_config_manager.set_default_lean_config_path(value)

    def _parse_verbose_option(self, ctx: click.Context, param: click.Parameter, value: Optional[bool]) -> None:
        """Parses the --verbose option."""
        if value:
            logger = container.logger()
            logger.debug_logging_enabled = True


class PathParameter(click.ParamType):
    """A limited version of click.Path which uses pathlib.Path."""

    def __init__(self, exists: bool = False, file_okay: bool = True, dir_okay: bool = True):
        """Creates a new PathParameter instance.

        :param exists: True if the path needs to point to an existing object, False if not
        :param file_okay: True if the path may point to a file, False if not
        :param dir_okay: True if the path may point to a directory, False if not
        """
        self._exists = exists
        self._file_okay = file_okay
        self._dir_okay = dir_okay

        if file_okay and not dir_okay:
            self.name = "file"
            self._path_type = "File"
        elif dir_okay and not file_okay:
            self.name = "directory"
            self._path_type = "Directory"
        else:
            self.name = "path"
            self._path_type = "Path"

    def convert(self, value: str, param: click.Parameter, ctx: click.Context) -> Path:
        path = Path(value).expanduser().resolve()

        if not container.path_manager().is_path_valid(path):
            self.fail(f"{self._path_type} '{value}' is not a valid path.", param, ctx)

        if self._exists and not path.exists():
            self.fail(f"{self._path_type} '{value}' does not exist.", param, ctx)

        if not self._file_okay and path.is_file():
            self.fail(f"{self._path_type} '{value}' is a file.", param, ctx)

        if not self._dir_okay and path.is_dir():
            self.fail(f"{self._path_type} '{value}' is a directory.", param, ctx)

        return path


class DateParameter(click.ParamType):
    """A click parameter which returns datetime.datetime objects and requires yyyyMMdd input."""

    name = "date"

    def get_metavar(self, param: click.Parameter) -> str:
        return "[yyyyMMdd]"

    def convert(self, value: str, param: click.Parameter, ctx: click.Context) -> datetime:
        for date_format in ["%Y%m%d", "%Y-%m-%d"]:
            try:
                return datetime.strptime(value, date_format)
            except ValueError:
                pass

        self.fail(f"'{value}' does not match the yyyyMMdd format.", param, ctx)


def ensure_options(options: List[str]) -> None:
    """Ensures certain options have values, raises an error if not.

    :param options: the Python names of the options that must have values
    """
    ctx = click.get_current_context()

    missing_options = []
    for key, value in ctx.params.items():
        has_value = value is not None

        if isinstance(value, tuple) and len(value) == 0:
            has_value = False

        if not has_value and key in options:
            missing_options.append(key)

    if len(missing_options) == 0:
        return

    missing_options = sorted(missing_options, key=lambda param: options.index(param))
    help_records = []

    for name in missing_options:
        option = next(param for param in ctx.command.params if param.name == name)
        help_records.append(option.get_help_record(ctx))

    help_formatter = click.HelpFormatter(max_width=120)
    help_formatter.write_dl(help_records)

    raise RuntimeError(f"""
You are missing the following option{"s" if len(missing_options) > 1 else ""}:
{''.join(help_formatter.buffer)}
    """.strip())



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import platform


class PlatformManager:
    """The PlatformManager class makes it easy to detect which platform the user is using."""

    def __init__(self) -> None:
        """Creates a new PlatformManager instance."""
        self._system = platform.system()
        self._machine = platform.machine()
        self._host_system = os.environ.get("QC_DOCKER_HOST_SYSTEM", None)
        self._host_machine = os.environ.get("QC_DOCKER_HOST_MACHINE", None)

    def is_system_windows(self) -> bool:
        """Returns whether the current system is running Windows.

        If this method is called inside Docker, it returns whether the Docker container is running Windows.

        :return: whether the current system is running Windows
        """
        return self._system == "Windows"

    def is_system_macos(self) -> bool:
        """Returns whether the current system is running macOS.

        If this method is called inside Docker, it returns whether the Docker container is running macOS.

        :return: whether the current system is running macOS
        """
        return self._system == "Darwin"

    def is_system_linux(self) -> bool:
        """Returns whether the current system is running Linux.

        If this method is called inside Docker, it returns whether the Docker container is running Linux.

        :return: whether the current system is running Linux
        """
        return self._system == "Linux"

    def is_system_arm(self) -> bool:
        """Returns whether the current system is running on the ARM architecture.

        If this method is called inside Docker, it returns whether the Docker container is running on ARM.

        :return: whether the current system is running on ARM
        """
        return self._machine in ["arm64", "aarch64"]

    def is_host_windows(self) -> bool:
        """Returns whether the current host is running Windows.

        If this method is called outside Docker, it behaves identical to is_system_windows().
        If this method is called inside Docker, it returns whether the Docker host is running Windows.

        :return: whether the current host is running Windows
        """
        if self._host_system is None:
            return self.is_system_windows()
        return self._host_system == "Windows"

    def is_host_macos(self) -> bool:
        """Returns whether the current host is running macOS.

        If this method is called outside Docker, it behaves identical to is_system_macos().
        If this method is called inside Docker, it returns whether the Docker host is running macOS.

        :return: whether the current host is running macOS
        """
        if self._host_system is None:
            return self.is_system_macos()
        return self._host_system == "Darwin"

    def is_host_linux(self) -> bool:
        """Returns whether the current host is running Linux.

        If this method is called outside Docker, it behaves identical to is_system_linux().
        If this method is called inside Docker, it returns whether the Docker host is running Linux.

        :return: whether the current host is running Linux
        """
        if self._host_system is None:
            return self.is_system_linux()
        return self._host_system == "Linux"

    def is_host_arm(self) -> bool:
        """Returns whether the current host is running on the ARM architecture.

        If this method is called outside Docker, it behaves identical to is_system_arm().
        If this method is called inside Docker, it returns whether the Docker host is running on ARM.

        :return: whether the current host is running on ARM
        """
        if self._host_machine is None:
            return self.is_system_arm()
        return self._host_machine in ["arm64", "aarch64"]



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List

import click

from lean.components.api.api_client import APIClient
from lean.components.util.logger import Logger
from lean.components.util.task_manager import TaskManager
from lean.models.api import QCBacktest, QCCompileState, QCCompileWithLogs, QCOptimization, QCProject
from lean.models.errors import RequestFailedError
from lean.models.optimizer import OptimizationConstraint, OptimizationParameter, OptimizationTarget


class CloudRunner:
    """The CloudRunner is responsible for running projects in the cloud."""

    def __init__(self, logger: Logger, api_client: APIClient, task_manager: TaskManager):
        """Creates a new CloudBacktestRunner instance.

        :param logger: the logger to use to log messages with
        :param api_client: the APIClient instance to use when communicating with the QuantConnect API
        :param task_manager: the TaskManager to run long-running tasks with
        """
        self._logger = logger
        self._api_client = api_client
        self._task_manager = task_manager

    def run_backtest(self, project: QCProject, name: str) -> QCBacktest:
        """Runs a backtest in the cloud.

        :param project: the project to backtest
        :param name: the name of the backtest
        :return: the completed backtest
        """
        finished_compile = self.compile_project(project)
        created_backtest = self._api_client.backtests.create(project.projectId, finished_compile.compileId, name)

        self._logger.info(f"Started backtest named '{name}' for project '{project.name}'")
        self._logger.info(f"Backtest url: {created_backtest.get_url()}")

        try:
            return self._task_manager.poll(
                make_request=lambda: self._api_client.backtests.get(project.projectId, created_backtest.backtestId),
                is_done=lambda data: data.is_complete(),
                get_progress=lambda data: data.progress
            )
        except KeyboardInterrupt as e:
            if click.confirm("Do you want to cancel and delete the running backtest?", True):
                self._api_client.backtests.delete(project.projectId, created_backtest.backtestId)
                self._logger.info(f"Successfully cancelled and deleted backtest '{name}'")
            raise e

    def run_optimization(self,
                         project: QCProject,
                         finished_compile: QCCompileWithLogs,
                         name: str,
                         strategy: str,
                         target: OptimizationTarget,
                         parameters: List[OptimizationParameter],
                         constraints: List[OptimizationConstraint],
                         node_type: str,
                         parallel_nodes: int) -> QCOptimization:
        """Runs an optimization in the cloud.

        :param project: the project to optimize
        :param finished_compile: a finished compile of the given project
        :param name: the name of the optimization
        :param strategy: the strategy to optimize with
        :param target: the target of the optimization
        :param parameters: the parameters to optimize
        :param constraints: the constraints of the optimization
        :param node_type: the type of the node to run the optimization on
        :param parallel_nodes: the number of parallel nodes to run the optimization on
        :return: the completed optimization
        """
        created_optimization = self._api_client.optimizations.create(project.projectId,
                                                                     finished_compile.compileId,
                                                                     name,
                                                                     strategy,
                                                                     target,
                                                                     parameters,
                                                                     constraints,
                                                                     node_type,
                                                                     parallel_nodes)

        self._logger.info(f"Started optimization named '{name}' for project '{project.name}'")
        self._logger.info(f"Project url: {project.get_url()}")

        try:
            return self._task_manager.poll(
                make_request=lambda: self._api_client.optimizations.get(created_optimization.optimizationId),
                is_done=lambda data: data.status != "active" and data.status != "running",
                get_progress=lambda data: data.get_progress()
            )
        except KeyboardInterrupt as e:
            if click.confirm("Do you want to cancel and delete the running optimization?", True):
                try:
                    self._api_client.optimizations.abort(created_optimization.optimizationId)
                except RequestFailedError:
                    # The optimization finished between the user pressed Ctrl+C and confirmed the deletion
                    pass
                finally:
                    self._api_client.optimizations.delete(created_optimization.optimizationId)
                    self._logger.info(f"Successfully cancelled and deleted optimization '{name}'")
            raise e

    def compile_project(self, project: QCProject) -> QCCompileWithLogs:
        """Compiles a project in the cloud.

        :param project: the project to compile
        :return: a QCCompileWithLogs instance containing the details of the finished compile
        """
        self._logger.info(f"Started compiling project '{project.name}'")

        created_compile = self._api_client.compiles.create(project.projectId)

        # Log the parameters reported in the compile
        parameters = []
        parameter_count = 0

        for parameter_container in created_compile.parameters:
            for parameter in parameter_container.parameters:
                parameters.append(f"- {parameter_container.file}:{parameter.line} :: {parameter.type}")
                parameter_count += int(parameter.type.split(" ")[0])

        if parameter_count > 0:
            self._logger.info(f"Detected parameters ({parameter_count}):")
            for parameter in parameters:
                self._logger.info(parameter)
        else:
            self._logger.info("Detected parameters: none")

        finished_compile = self._task_manager.poll(
            make_request=lambda: self._api_client.compiles.get(project.projectId, created_compile.compileId),
            is_done=lambda data: data.state in [QCCompileState.BuildSuccess, QCCompileState.BuildError]
        )

        if finished_compile.state == QCCompileState.BuildError:
            self._logger.error("\n".join(finished_compile.logs))
            raise RuntimeError(f"Something went wrong while compiling project '{project.name}'")

        self._logger.info("\n".join(finished_compile.logs))
        self._logger.info(f"Successfully compiled project '{project.name}'")

        return finished_compile



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import click

class DefaultCommandGroup(click.Group):
    """allow a default command for a group"""

    def command(self, *args, **kwargs):
        default_command = kwargs.pop('default_command', False)
        if default_command and not args:
            kwargs['name'] = kwargs.get('name', '<>')
        decorator = super(
            DefaultCommandGroup, self).command(*args, **kwargs)

        if default_command:
            def new_decorator(f):
                cmd = decorator(f)
                self.default_command = cmd.name
                return cmd

            return new_decorator

        return decorator

    def resolve_command(self, ctx, args):
        try:
            # test if the command parses
            return super(
                DefaultCommandGroup, self).resolve_command(ctx, args)
        except click.UsageError:
            # command did not parse, assume it is the default command
            args.insert(0, self.default_command)
            return super(
                DefaultCommandGroup, self).resolve_command(ctx, args)


# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import click

from lean.commands.library.add import add
from lean.commands.library.remove import remove


@click.group()
def library() -> None:
    """Manage custom libraries in a project."""
    # This method is intentionally empty
    # It is used as the command group for all `lean library <command>` commands
    pass


library.add_command(add)
library.add_command(remove)



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path

import click

from lean.click import LeanCommand
from lean.commands import lean
from lean.container import container


@lean.command(cls=LeanCommand, name="project-delete", aliases=["delete-project"])
@click.argument("project", type=str)
def delete_project(project: str) -> None:
    """Delete a project locally and in the cloud if it exists.

    The project is selected by name or cloud id.
    """
    # Remove project from cloud
    api_client = container.api_client()
    all_projects = api_client.projects.get_all()
    project_manager = container.project_manager()
    logger = container.logger()

    project_id = None
    try:
        project_id = int(project)
    except ValueError:
        pass

    projects = []
    try:
        projects = project_manager.get_projects_by_name_or_id(all_projects, project_id or project)
    except RuntimeError:
        # If searching by id, we cannot try lo locate the project locally
        if project_id is not None:
            raise RuntimeError(f"The project with ID {project_id} was not found in the cloud.")

        # The project might only be local, look for the path
        logger.info(f"The project {project} was not found in the cloud. It will be removed locally if it exists.")

    full_project = next(iter(projects), None)

    if full_project is not None:
        api_client.projects.delete(full_project.projectId)

    # Remove project locally
    project_path = full_project.name if full_project is not None else project
    project_manager.delete_project(Path(project_path))

    logger.info(f"Successfully deleted project '{project_path}'")



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import shutil
import tempfile
from pathlib import Path


class TempManager:
    """The TempManager class provides access to temporary directories."""

    def __init__(self) -> None:
        """Creates a new TempManager instance."""
        self._temporary_directories = []
        self.delete_temporary_directories_when_done = True

    def create_temporary_directory(self) -> Path:
        """Returns the path to an empty temporary directory.

        :return: a path to an empty temporary directory
        """
        path = Path(tempfile.mkdtemp(prefix="lean-cli-"))
        self._temporary_directories.append(path)
        return path

    def delete_temporary_directories(self) -> None:
        """Deletes temporary directories that were created while the CLI ran.

        Only the files that the user can delete are deleted, any permission errors are ignored.
        """
        for path in self._temporary_directories:
            shutil.rmtree(path, ignore_errors=True)



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Optional

from lean.components.api.api_client import *
from lean.models.api import QCLanguage, QCProject


class ProjectClient:
    """The ProjectClient class contains methods to interact with projects/* API endpoints."""

    def __init__(self, api_client: 'APIClient') -> None:
        """Creates a new ProjectClient instance.

        :param api_client: the APIClient instance to use when making requests
        """
        self._api = api_client

    def get(self, project_id: int) -> QCProject:
        """Returns the details of a project.

        :param project_id: the id of the project to retrieve the details of
        :return: the details of the specified project
        """
        data = self._api.get("projects/read", {
            "projectId": project_id
        })

        return self._process_project(QCProject(**data["projects"][0]))

    def get_all(self) -> List[QCProject]:
        """Returns all the projects the user has access to.

        :return: a list containing all the projects the user has access to
        """
        data = self._api.get("projects/read")
        return [self._process_project(QCProject(**project)) for project in data["projects"]]

    def create(self, name: str, language: QCLanguage, organization_id: Optional[str]) -> QCProject:
        """Creates a new project.

        :param name: the name of the project to create
        :param language: the language of the project to create
        :param organization_id: the id of the organization to create the project in
        :return: the created project
        """
        parameters = {
            "name": name,
            "language": language.value
        }
        if organization_id is not None:
            parameters["organizationId"] = organization_id
        data = self._api.post("projects/create", parameters)

        return self._process_project(QCProject(**data["projects"][0]))

    def update(self,
               project_id: int,
               name: Optional[str] = None,
               description: Optional[str] = None,
               parameters: Optional[Dict[str, str]] = None,
               lean_engine: Optional[int] = None,
               python_venv: Optional[int] = None) -> None:
        """Updates an existing project.

        :param project_id: the id of the project to update
        :param name: the new name to assign to the project, or None if the name shouldn't be changed
        :param description: the new description to assign to the project, or None if the description shouldn't be changed
        :param parameters: the new parameters of the project, or None if the parameters shouldn't be changed
        :param lean_engine: the lean engine id for the project, or None if the lean engine shouldn't be changed
        :param python_venv: the python venv id for the project, or None if the python venv shouldn't be changed
        """
        request_parameters = {
            "projectId": project_id
        }

        if name is not None:
            request_parameters["name"] = name

        if description is not None:
            request_parameters["description"] = description

        if parameters is not None:
            if len(parameters) > 0:
                index = 0
                for key, value in parameters.items():
                    request_parameters[f"parameters[{index}][key]"] = key
                    request_parameters[f"parameters[{index}][value]"] = value
                    index += 1
            else:
                request_parameters["parameters"] = ""

        if lean_engine is not None:
            request_parameters["versionId"] = lean_engine

        if python_venv is not None:
            request_parameters["leanEnvironment"] = python_venv

        self._api.post("projects/update", request_parameters, data_as_json=False)

    def delete(self, project_id: int) -> None:
        """Deletes an existing project.

        :param project_id: the id of the project to delete
        """
        self._api.post("projects/delete", {
            "projectId": project_id
        })

    def add_library(self, project_id: int, library_project_id: int) -> None:
        """Links a library to a project.

        :param project_id: the id of the project to add the library to
        :param library_project_id: the id of the library project to add as library
        """
        self._api.post("projects/library/create", {
            "projectId": project_id,
            "libraryId": library_project_id
        })

    def delete_library(self, project_id: int, library_project_id: int) -> None:
        """Removes a library from a project.

        :param project_id: the id of the project the library is added to
        :param library_project_id: the id of the library project to remove
        """
        self._api.post("projects/library/delete", {
            "projectId": project_id,
            "libraryId": library_project_id
        })

    def _process_project(self, project: Any) -> Any:
        """Patches a cloud project's name by ensuring the name doesn't start with a slash.

        :param project: the project to process
        :return: the project that is given after patching it
        """
        project.name = project.name.lstrip("/")
        return project



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from pathlib import Path
from typing import Any


class Storage:
    """A Storage instance manages the data in a single JSON file."""

    def __init__(self, file: str) -> None:
        """Creates a new Storage instance.

        :param file: the path to the file this Storage instance should manage
        """
        self.file = Path(file)

        if self.file.exists():
            self._data = json.loads(self.file.read_text(encoding="utf-8"))
        else:
            self._data = {}

    def get(self, key: str, default: Any = None) -> Any:
        """Returns the value assigned to the given key.

        Returns a default value when nothing is assigned to the given key.

        :param key: the key to retrieve the value of
        :param default: the default value to return when key is not set
        :return: the value assigned to the key or default if the key is not set
        """
        if key in self._data:
            return self._data[key]
        else:
            return default

    def set(self, key: str, value: Any) -> None:
        """Assigns a value to a key.

        The value is stored as json, so must be serializable using json.dump().

        :param key: the key to assign the value to
        :param value: the json-serializable value to assign to the given key
        """
        self._data[key] = value
        self._save()

    def delete(self, key: str) -> None:
        """Deletes a key.

        :param key: the key to delete
        """
        self._data.pop(key, None)
        self._save()

    def has(self, key: str) -> bool:
        """Returns whether the Storage instance has a value assigned to the given key.

        :param key: the key to check the existence of
        :return: True if a value is assigned to the given key, False if not
        """
        return key in self._data

    def clear(self) -> None:
        """Clears the Storage instance and deletes the underlying file."""
        self._data.clear()
        self._save()

    def _save(self) -> None:
        """Saves the data to the underlying file, deleting the file if there is no data."""
        if len(self._data) > 0:
            self.file.parent.mkdir(parents=True, exist_ok=True)

            with self.file.open("w+", encoding="utf-8") as file:
                file.write(json.dumps(self._data, indent=4) + "\n")
        else:
            if self.file.exists():
                self.file.unlink()



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from pathlib import Path
import uuid
import click
from lean.click import LeanCommand, PathParameter
from typing import Optional
from lean.commands.live.live import get_result, send_command
from lean.components.util.click_custom_parameters import DECIMAL

@click.command(cls=LeanCommand, requires_lean_config=True, requires_docker=True)
@click.argument("project", type=PathParameter(exists=True, file_okay=True, dir_okay=True))
@click.option("--order-id",
              type=int,
              required=True,
              help="The order id to be updated")
@click.option("--quantity", type=DECIMAL, help="The number of units to be updated (directional)")
@click.option("--limit-price", type=DECIMAL, help="The limit price of the order to be updated")
@click.option("--stop-price", type=DECIMAL, help="The stop price of the order to be updated")
@click.option("--tag", type=str, help="The tag to be attached to the order")
def update_order(project: Path,
                 order_id: str,
                 quantity: Optional[DECIMAL],
                 limit_price: Optional[DECIMAL],
                 stop_price: Optional[DECIMAL],
                 tag: Optional[str]) -> None:
    """
    Represents a command to update a specific order by id.
    """

    command_id = uuid.uuid4().hex

    data = {
        "$type": "QuantConnect.Commands.UpdateOrderCommand, QuantConnect.Common",
        "Id": command_id,
        "OrderId": order_id,
        "Quantity": quantity,
        "LimitPrice": limit_price,
        "StopPrice": stop_price,
        "Tag": tag
    }

    docker_container_name = send_command(project, data)
    get_result(command_id, docker_container_name)



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import shutil
import zipfile
from pathlib import Path

import click

from lean.click import LeanCommand
from lean.constants import DEFAULT_DATA_DIRECTORY_NAME, DEFAULT_LEAN_CONFIG_FILE_NAME
from lean.container import container
from lean.models.errors import MoreInfoError


def _download_repository(output_path: Path) -> None:
    """Downloads the LEAN repository as a zip file.

    :param output_path: the path to the zip file where the LEAN repository must be saved to
    """
    logger = container.logger()
    logger.info("Downloading latest sample data from the Lean repository...")

    # We download the entire Lean repository and extract the data and the launcher's config file
    # GitHub doesn't allow downloading a specific directory
    # Since we need ~80% of the total repository in terms of file size this shouldn't be too big of a problem
    response = container.http_client().get("https://github.com/QuantConnect/Lean/archive/master.zip", stream=True)

    total_size_bytes = int(response.headers.get("content-length", 0))
    print_file_size_at_mb = 10

    # Sometimes content length isn't set, don't show a progress bar in that case
    if total_size_bytes > 0:
        progress = logger.progress()
        progress_task = progress.add_task("")
    else:
        progress = progress_task = None

    try:
        with output_path.open("wb") as file:
            written_bytes = 0

            for chunk in response.iter_content(chunk_size=8192):
                file.write(chunk)
                written_bytes += len(chunk)
                
                if progress is not None:
                    progress.update(progress_task, completed=(written_bytes / total_size_bytes) * 100)
                
                file_size_mb = round(written_bytes / 1024 / 1024, 2)
                if file_size_mb >= print_file_size_at_mb:
                    print_file_size_at_mb += 20
                    logger.info(f"Download Progress {file_size_mb} MB")
    except KeyboardInterrupt as e:
        if progress is not None:
            progress.stop()
        raise e

    if progress is not None:
        progress.stop()


@click.command(cls=LeanCommand)
@click.option("--language", "-l",
              type=click.Choice(container.cli_config_manager().default_language.allowed_values, case_sensitive=False),
              help="The default language to use for new projects")
def init(language: str) -> None:
    """Scaffold a Lean configuration file and data directory."""
    current_dir = Path.cwd()
    data_dir = current_dir / DEFAULT_DATA_DIRECTORY_NAME
    lean_config_path = current_dir / DEFAULT_LEAN_CONFIG_FILE_NAME

    # Abort if one of the files we are going to create already exists to prevent us from overriding existing files
    for path in [data_dir, lean_config_path]:
        if path.exists():
            relative_path = path.relative_to(current_dir)
            raise MoreInfoError(f"{relative_path} already exists, please run this command in an empty directory",
                                "https://www.lean.io/docs/v2/lean-cli/initialization/directory-structure#02-lean-init")

    logger = container.logger()

    # Warn the user if the current directory is not empty
    if next(current_dir.iterdir(), None) is not None:
        logger.info("This command will create a Lean configuration file and data directory in the current directory")
        click.confirm("The current directory is not empty, continue?", default=False, abort=True)

    # Download the Lean repository
    tmp_directory = container.temp_manager().create_temporary_directory()
    _download_repository(tmp_directory / "master.zip")

    # Extract the downloaded repository
    with zipfile.ZipFile(tmp_directory / "master.zip") as zip_file:
        zip_file.extractall(tmp_directory / "master")

    # Copy the data directory
    shutil.copytree(tmp_directory / "master" / "Lean-master" / "Data", data_dir)

    # Create the config file
    lean_config_manager = container.lean_config_manager()
    config = (tmp_directory / "master" / "Lean-master" / "Launcher" / "config.json").read_text(encoding="utf-8")
    config = lean_config_manager.clean_lean_config(config)
    lean_config_manager.store_known_lean_config_path(lean_config_path)

    # Update the data-folder configuration
    config = config.replace('"data-folder": "../../../Data/"', f'"data-folder": "{DEFAULT_DATA_DIRECTORY_NAME}"')

    with lean_config_path.open("w+", encoding="utf-8") as file:
        file.write(config)

    # Prompt for some general configuration if not set yet
    cli_config_manager = container.cli_config_manager()
    if cli_config_manager.default_language.get_value() is None:
        default_language = language if language is not None else click.prompt("What should the default language for new projects be?",
                                        default=cli_config_manager.default_language.default_value,
                                        type=click.Choice(cli_config_manager.default_language.allowed_values))
        cli_config_manager.default_language.set_value(default_language)

    logger.info(f"""
The following objects have been created:
- {DEFAULT_LEAN_CONFIG_FILE_NAME} contains the configuration used when running the LEAN engine locally
- {DEFAULT_DATA_DIRECTORY_NAME}/ contains the data that is used when running the LEAN engine locally

The following documentation pages may be useful:
- Setting up local autocomplete: https://www.lean.io/docs/v2/lean-cli/projects/autocomplete
- Synchronizing projects with the cloud: https://www.lean.io/docs/v2/lean-cli/projects/cloud-synchronization

Here are some commands to get you going:
- Run `lean create-project "My Project"` to create a new project with starter code
- Run `lean cloud pull` to download all your QuantConnect projects to your local drive
- Run `lean backtest "My Project"` to backtest a project locally with the data in {DEFAULT_DATA_DIRECTORY_NAME}/
""".strip())



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import multiprocessing
import tarfile
from pathlib import Path
from datetime import *
from typing import Any, List, Callable

from joblib import delayed, Parallel

from lean.components.api.api_client import APIClient
from lean.components.config.lean_config_manager import LeanConfigManager
from lean.components.util.logger import Logger
from lean.models.errors import MoreInfoError, RequestFailedError


def _store_local_file(file_content: bytes, file_path: Path):
    file_path.parent.mkdir(parents=True, exist_ok=True)
    with file_path.open("wb+") as f:
        f.write(file_content)


class DataDownloader:
    """The DataDownloader is responsible for downloading data from QuantConnect Datasets."""

    def __init__(self, logger: Logger, api_client: APIClient, lean_config_manager: LeanConfigManager):
        """Creates a new CloudBacktestRunner instance.

        :param logger: the logger to use to log messages with
        :param api_client: the APIClient instance to use when communicating with the QuantConnect API
        :param lean_config_manager: the LeanConfigManager instance to retrieve the data directory from
        """
        self._logger = logger
        self._api_client = api_client
        self._lean_config_manager = lean_config_manager

    def update_database_files(self):
        """Will update lean data folder database files if required

        """
        try:
            now = datetime.now()
            config = self._lean_config_manager.get_lean_config()
            last_update = config["file-database-last-update"] if "file-database-last-update" in config else ''
            if not last_update or now - datetime.strptime(last_update, '%m/%d/%Y') > timedelta(days=1):
                data_dir = self._lean_config_manager.get_data_directory()

                _store_local_file(self._api_client.data.download_public_file(
                    "https://raw.githubusercontent.com/QuantConnect/Lean/master/Data/symbol-properties/symbol-properties-database.csv"),
                    data_dir / "symbol-properties" / "symbol-properties-database.csv")
                _store_local_file(self._api_client.data.download_public_file(
                    "https://raw.githubusercontent.com/QuantConnect/Lean/master/Data/market-hours/market-hours-database.json"),
                    data_dir / "market-hours" / "market-hours-database.json")

                self._lean_config_manager.set_properties({"file-database-last-update": now.strftime('%m/%d/%Y')})
        except MoreInfoError as e:
            if "not found" in str(e):
                pass
            else:
                self._logger.error(str(e))
        except Exception as e:
            self._logger.error(str(e))

    def download_files(self, data_files: List[Any], overwrite: bool, organization_id: str) -> None:
        """Downloads files from QuantConnect Datasets to the local data directory.

        :param data_files: the list of data files to download
        :param overwrite: whether existing files may be overwritten
        :param organization_id: the id of the organization that should be billed
        """
        progress = self._logger.progress(suffix="{task.percentage:0.0f}% ({task.completed:,.0f}/{task.total:,.0f})")
        progress_task = progress.add_task("", total=len(data_files))

        try:
            parallel = Parallel(n_jobs=max(1, multiprocessing.cpu_count() - 1), backend="threading")

            data_dir = self._lean_config_manager.get_data_directory()
            parallel(delayed(self._download_file)(data_file.file, overwrite, data_dir, organization_id,
                                                  lambda advance: progress.update(progress_task, advance=advance))
                     for data_file in data_files)

            # update our config after we download all files, and not in parallel!
            for datafile in data_files:
                relative_file = datafile.file
                if "/map_files/map_files_" in relative_file and relative_file.endswith(".zip"):
                    self._lean_config_manager.set_properties({
                        "map-file-provider": "QuantConnect.Data.Auxiliary.LocalZipMapFileProvider"
                    })
                if "/factor_files/factor_files_" in relative_file and relative_file.endswith(".zip"):
                    self._lean_config_manager.set_properties({
                        "factor-file-provider": "QuantConnect.Data.Auxiliary.LocalZipFactorFileProvider"
                    })

            progress.stop()
        except KeyboardInterrupt as e:
            progress.stop()
            raise e

    def _process_bulk(self, file: Path, destination: Path):
        tar = tarfile.open(file)
        tar.errorlevel = 0
        tar.extractall(destination)
        tar.close()

    def _download_file(self,
                       relative_file: str,
                       overwrite: bool,
                       data_directory: Path,
                       organization_id: str,
                       progress_callback: Callable[[float], None]) -> None:
        """Downloads a single file from QuantConnect Datasets to the local data directory.

        If this method downloads a map or factor files zip file,
        it also updates the Lean config file to ensure LEAN uses those files instead of the csv files.

        :param relative_file: the relative path to the file in the data directory
        :param overwrite: whether existing files may be overwritten
        :param data_directory: the path to the local data directory
        :param organization_id: the id of the organization that should be billed
        :param callback: the lambda that is called just before the method returns
        """
        local_path = data_directory / relative_file

        if local_path.exists() and not overwrite:
            self._logger.warn("\n".join([
                f"{local_path} already exists, use --overwrite to overwrite it",
                "You have not been charged for this file"
            ]))
            progress_callback(1)
            return

        try:
            self._api_client.data.download_file(relative_file, organization_id, local_path, progress_callback)
        except RequestFailedError as error:
            self._logger.warn(f"{local_path}: {error}\nYou have not been charged for this file")
            progress_callback(1)
            return

        # Special case: bulk files need unpacked
        if "setup/" in relative_file and relative_file.endswith(".tar"):
            self._process_bulk(local_path, data_directory)



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from typing import Dict, Any
from lean.container import container
import sys
import io
from contextlib import redirect_stdout
import re

docker_manager = container.docker_manager()
project_manager = container.project_manager()
lean_runner = container.lean_runner()
temp_manager = container.temp_manager()
project_config_manager = container.project_config_manager()
cli_config_manager = container.cli_config_manager()
logger = container.logger()


def get_success() -> Dict[str, Any]:
    """Compiles success message.

    :return: success object as json dump.
    """
    return json.dumps({
        "eType": "BuildSuccess",
    })


def get_errors(algorithm_type: str, message: str, color_coding_required: bool = True,
                                        warning_required: bool = True) -> Dict[str, Any]:
    """Compiles error message based on given input

    :param algorithm_type: type of algorithm: "python" or "csharp".
    :param message: message from which errors needs to be formatted.
    :return: error object as json dump.
    """
    errors = []
    if algorithm_type == "csharp":
        errors.extend(_parse_csharp_errors(message, color_coding_required, warning_required))
    elif algorithm_type == "python":
        errors.extend(_parse_python_errors(message, color_coding_required))

    return json.dumps({
        "eType": "BuildError",
        "aErrors": errors,
    })


def redirect_stdout_of_subprocess(method_name_to_run, *args, **kwargs) -> tuple:
    """ It captures the stdout of the method given to run.

    :param method_name_to_run: name of the method to run
    :return: result of the method and the stdout of the process
    """
    f = io.StringIO()
    with redirect_stdout(f):
        result = method_name_to_run(*args, **kwargs)
    stdout = f.getvalue()
    return (result, stdout)


def compile() -> None:
    """This is a utility function that is used by the vscode plugin project.
    """

    # We need to print the stdout of the docker run command from here,
    # so that it can be picked up by the subprocess that is being 
    # called by the vscode plugin.
    compile_result, stdout = redirect_stdout_of_subprocess(_compile)
    if compile_result["result"]:
        processed_output = get_success()
    else:
        processed_output = get_errors(compile_result["algorithmType"], stdout, False, False)
    logger.info(processed_output)


def _compile() -> Dict[str, Any]:
    """
    This function compile c# and python project files.
    """
    message = {
        "result": False,
        "algorithmType": "",
    }

    project_id = int(sys.argv[-1])
    project_dir = project_manager.get_project_by_id(project_id)
    algorithm_file = project_manager.find_algorithm_file(project_dir)
    message["algorithmType"] = "python" if algorithm_file.name.endswith(".py") else "csharp"

    # The dict containing all options passed to `docker run`
    # See all available options at https://docker-py.readthedocs.io/en/stable/containers.html
    run_options: Dict[str, Any] = {
        "commands": [],
        "environment": {},
        "mounts": [],
        "volumes": {}
    }

    lean_runner.setup_language_specific_run_options(run_options, project_dir, algorithm_file, False, False)

    project_config = project_config_manager.get_project_config(project_dir)
    engine_image = cli_config_manager.get_engine_image(
        project_config.get("engine-image", None))

    message["result"] = docker_manager.run_image(engine_image, **run_options)
    temp_manager.delete_temporary_directories_when_done = False
    return message

def _parse_csharp_errors(csharp_output: str, color_coding_required: bool, warning_required: bool) -> list:
    errors = []

    try:
        relevant_output = csharp_output[csharp_output.index("Build FAILED."):]
        for match in re.findall(r"(.*)\((\d+),(\d+)\): (error|warning) ([a-zA-Z0-9]+): ([^[]+) ", relevant_output):
            if color_coding_required:
                if match[3] == "error":
                    errors.append(f'{bcolors.FAIL}{match[3]} File: {match[0].split("/")[-1]} Line {match[1]} Column {match[2]} - {match[5]}{bcolors.ENDC}\n')
                elif warning_required:
                    errors.append(f'{bcolors.WARNING}{match[3]}: {match[0].split("/")[-1]} Line {match[1]} Column {match[2]} - {match[5]}{bcolors.ENDC}\n')
            else:
                if match[3] == "warning" and not warning_required:
                    continue
                errors.append(f'{match[3]}: {match[0].split("/")[-1]} Line {match[1]} Column {match[2]} - {match[5]}\n')
    except Exception:
        pass

    return errors

def _parse_python_errors(python_output: str, color_coding_required: bool) -> list:
    errors = []

    try:
        for match in re.findall(r'\*\*\*   File "/LeanCLI/([^"]+)", line (\d+)\n.*\n(.*)\^.*\n(.*)', python_output):
            if color_coding_required:
                errors.append(f"{bcolors.FAIL}Build Error File: {match[0]} Line {match[1]} Column {match[2]} - {match[3]}{bcolors.ENDC}\n")
            else:
                errors.append(f"Build Error File: {match[0]} Line {match[1]} Column {match[2]} - {match[3]}\n")

        for match in re.findall(r"\*\*\* Sorry: ([^(]+) \(([^,]+), line (\d+)\)", python_output):
            if color_coding_required:
                errors.append(f"{bcolors.FAIL}Build Error File: {match[1]} Line {match[2]} Column 0 - {match[0]}{bcolors.ENDC}\n")
            else:
                errors.append(f"Build Error File: {match[1]} Line {match[2]} Column 0 - {match[0]}\n")
    except Exception:
        pass
    
    return errors


class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    OKGREEN = '\033[92m'
    WARNING = '\033[33m'
    FAIL = '\033[31m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import click

from lean.commands.cloud.backtest import backtest
from lean.commands.cloud.live.live import live
from lean.commands.cloud.optimize import optimize
from lean.commands.cloud.pull import pull
from lean.commands.cloud.push import push
from lean.commands.cloud.status import status


@click.group()
def cloud() -> None:
    """Interact with the QuantConnect cloud."""
    # This method is intentionally empty
    # It is used as the command group for all `lean cloud <command>` commands
    pass


cloud.add_command(pull)
cloud.add_command(push)
cloud.add_command(backtest)
cloud.add_command(optimize)
cloud.add_command(live)
cloud.add_command(status)



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path
from typing import Set, List, Dict

from lean.components.api.api_client import APIClient
from lean.components.util.http_client import HTTPClient
from lean.components.util.logger import Logger
from lean.constants import MODULES_DIRECTORY
from lean.models.modules import NuGetPackage


class ModuleManager:
    """The ModuleManager class is responsible for downloading and updating modules."""

    def __init__(self, logger: Logger, api_client: APIClient, http_client: HTTPClient) -> None:
        """Creates a new ModuleManager instance.

        :param logger: the logger to use
        :param api_client: the APIClient instance to use when communicating with the cloud
        :param http_client: the HTTPClient instance to use when downloading modules
        """
        self._logger = logger
        self._api_client = api_client
        self._http_client = http_client
        self._installed_product_ids: Set[int] = set()
        self._installed_packages: Dict[int, List[NuGetPackage]] = {}

    def install_module(self, product_id: int, organization_id: str) -> None:
        """Installs a module into the global modules directory.

        If an outdated version is already installed, it is automatically updated.
        If the organization does not have a subscription for the given module, an error is raised.

        :param product_id: the product id of the module to download
        :param organization_id: the id of the organization that has a license for the module
        """
        if product_id in self._installed_product_ids:
            return

        module_files = self._api_client.modules.list_files(product_id, organization_id)
        packages_to_download: Dict[str, NuGetPackage] = {}

        for file_name in module_files:
            package = NuGetPackage.parse(file_name)

            if package.name not in packages_to_download or package.version > packages_to_download[package.name].version:
                packages_to_download[package.name] = package

        for package in packages_to_download.values():
            self._download_file(product_id, organization_id, package)

        self._installed_product_ids.add(product_id)

    def get_installed_packages(self) -> List[NuGetPackage]:
        """Returns a list of NuGet packages that were installed by install_module() calls.

        :return: a list of NuGet packages in the modules directory that should be made available when running LEAN
        """
        packages = []
        for package_list in self._installed_packages.values():
            packages.extend(package_list)
        return packages

    def get_installed_packages_by_module(self, product_id: int) -> List[NuGetPackage]:
        """Returns a list of NuGet packages that were installed by install_module() for a given product id.

        :param product_id: the product id to get the installed packages of
        :return: a list of NuGet packages in are available for the given product id
        """
        return self._installed_packages.get(product_id, []).copy()

    def is_module_installed(self, product_id: int) -> bool:
        """Returns whether a module with a given producti d has been installed with install_module().

        :param product_id: the product id to check the install status of
        :return: True if the product id has been registered with install_module(), False if not
        """
        return product_id in self._installed_product_ids

    def _download_file(self, product_id: int, organization_id: str, package: NuGetPackage) -> None:
        """Downloads a file if it doesn't already exist locally.

        :param product_id: the product id of the module to download
        :param organization_id: the id of the organization that has a license for the module
        :param package: the NuGet package to download
        """
        package_file = Path(MODULES_DIRECTORY) / package.get_file_name()

        if package_file.is_file():
            if product_id not in self._installed_packages:
                self._installed_packages[product_id] = []
            self._installed_packages[product_id].append(package)
            self._logger.debug(f"ModuleManager._download_file(): {package_file} already exists locally")
            return

        self._logger.info(f"Downloading '{package_file.name}'")

        package_file.parent.mkdir(parents=True, exist_ok=True)

        link = self._api_client.modules.get_link(product_id, organization_id, package_file.name)
        try:
            with self._http_client.get(link, stream=True) as response:
                with package_file.open("wb+") as file:
                    for chunk in response.iter_content(chunk_size=8192):
                        file.write(chunk)
        except Exception as exception:
            package_file.unlink(missing_ok=True)
            raise exception

        if product_id not in self._installed_packages:
            self._installed_packages[product_id] = []
        self._installed_packages[product_id].append(package)
        self._logger.debug(f"ModuleManager._download_file(): adding {package.name} to _installed_packages")


# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import hashlib
import os
from datetime import datetime, timedelta, timezone
from distutils.version import StrictVersion

import requests
from docker.errors import APIError
from rich import box
from rich.panel import Panel
from rich.table import Table

import lean
from lean.components.config.storage import Storage
from lean.components.docker.docker_manager import DockerManager
from lean.components.util.http_client import HTTPClient
from lean.components.util.logger import Logger
from lean.constants import (UPDATE_CHECK_INTERVAL_ANNOUNCEMENTS, UPDATE_CHECK_INTERVAL_CLI,
                            UPDATE_CHECK_INTERVAL_DOCKER_IMAGE)
from lean.models.docker import DockerImage


class UpdateManager:
    """The UpdateManager class contains methods to check for and warn the user about available updates."""

    def __init__(self,
                 logger: Logger,
                 http_client: HTTPClient,
                 cache_storage: Storage,
                 docker_manager: DockerManager) -> None:
        """Creates a new UpdateManager instance.

        :param logger: the logger to use when warning the user when something is outdated
        :param http_client: the HTTPClient instance to use for HTTP requests
        :param cache_storage: the Storage instance to use for getting/setting the last time a certain update check was performed
        :param docker_manager: the DockerManager instance to use to check for Docker updates
        """
        self._logger = logger
        self._http_client = http_client
        self._cache_storage = cache_storage
        self._docker_manager = docker_manager

    def warn_if_cli_outdated(self, force: bool = False) -> None:
        """Warns the user if the CLI is outdated.

        An update check is performed once every UPDATE_CHECK_INTERVAL_CLI hours, unless force is set to True.

        :param force: whether the update check interval should be bypassed (defaults to False)
        """
        current_version = lean.__version__

        # A development version is never considered outdated
        if current_version == "dev":
            return

        if not force and not self._should_check_for_updates("cli", UPDATE_CHECK_INTERVAL_CLI):
            return

        try:
            response = self._http_client.get("https://pypi.org/pypi/lean/json", raise_for_status=False)
        except requests.exceptions.ConnectionError:
            # The user may be offline, do nothing
            return

        if not response.ok:
            return

        latest_version = response.json()["info"]["version"]

        if StrictVersion(latest_version) > StrictVersion(current_version):
            self._logger.warn(f"A new release of the Lean CLI is available ({current_version} -> {latest_version})")
            self._logger.warn("Run `pip install --upgrade lean` to update to the latest version")

    def pull_docker_image_if_necessary(self, image: DockerImage, force: bool) -> None:
        """Pulls a Docker image if necessary.

        Docker images are pulled when they are not installed yet,
        and once every UPDATE_CHECK_INTERVAL_DOCKER_IMAGE hours (the interval is per image).

        :param image: the image to pull
        :param force: skip the interval check to force a pull
        """
        if not force and self._docker_manager.image_installed(image):
            if not self._should_check_for_updates(str(image), UPDATE_CHECK_INTERVAL_DOCKER_IMAGE):
                return

            local_digest = self._docker_manager.get_local_digest(image)

            # If the local digest does not exist the image was built locally so there is nothing to pull
            if local_digest is None:
                return

            try:
                remote_digest = self._docker_manager.get_remote_digest(image)
            except APIError:
                # The user may be offline, do nothing
                return

            # Do nothing if we're already using the latest image
            if remote_digest == local_digest:
                return

        self._docker_manager.pull_image(image)

    def show_announcements(self) -> None:
        """Shows the announcements if they have been updated.

        We check for new announcements once every UPDATE_CHECK_INTERVAL_ANNOUNCEMENTS hours.
        """
        if not self._should_check_for_updates("announcements", UPDATE_CHECK_INTERVAL_ANNOUNCEMENTS):
            return

        try:
            response = self._http_client.get(
                "https://raw.githubusercontent.com/QuantConnect/lean-cli/master/announcements.json",
                raise_for_status=False
            )
        except requests.exceptions.ConnectionError:
            # The user may be offline, do nothing
            return

        if not response.ok:
            return

        hash_cache_key = "last-announcements-hash"

        remote_hash = hashlib.md5(response.content).hexdigest()
        local_hash = self._cache_storage.get(hash_cache_key, None)

        if local_hash == remote_hash:
            return

        self._cache_storage.set(hash_cache_key, remote_hash)

        announcements = response.json()["announcements"]
        if len(announcements) == 0:
            return

        table = Table.grid(padding=(0, 1))
        table.add_column(overflow="fold")
        table.add_column(overflow="fold", ratio=1)

        for announcement in announcements:
            table.add_row(announcement["date"] + ":", announcement["message"])

        self._logger.info(Panel.fit(table, title="Announcements", box=box.SQUARE))

    def _should_check_for_updates(self, key: str, interval_hours: int) -> bool:
        """Returns whether an update check should be performed.

        When this method returns True, for the next <interval_hours> hours it returns False for <key>.

        :param key: the key which is used to identify the current update check
        :param interval_hours: the amount of hours between update checks for the given key
        :return: True if an update check should be performed, False if not
        """
        storage_key = f"last-update-check-{key}"
        should_check = False

        if not self._cache_storage.has(storage_key):
            # Perform an update check if we haven't ran one yet
            should_check = True
        else:
            last_update_check = datetime.fromtimestamp(self._cache_storage.get(storage_key), tz=timezone.utc)
            time_since_last_update_check = datetime.now(tz=timezone.utc) - last_update_check

            if time_since_last_update_check >= timedelta(hours=interval_hours):
                # Perform an update check if the time since the last update check is greater than the given interval
                should_check = True

        if should_check:
            # Save the current timestamp so for the next <interval_hours> hours we can return False for <key>
            self._cache_storage.set(storage_key, datetime.now(tz=timezone.utc).timestamp())

        return should_check



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path

from lean.components.util.platform_manager import PlatformManager


class PathManager:
    """The PathManager class provides utilities for working with paths."""

    def __init__(self, platform_manager: PlatformManager) -> None:
        """Creates a new PathManager instance.

        :param platform_manager: the PlatformManager used when checking which operating system is in use
        """
        self._platform_manager = platform_manager

    def get_relative_path(self, destination: Path, source: Path = Path.cwd()) -> Path:
        """Returns a path relative to another one.

        :param destination: the path to point to
        :param source: the root where the relative path is relative to
        :return: the destination path relative to the source path, or destination path if it is not relative
        """
        try:
            return destination.relative_to(source)
        except ValueError:
            return destination

    def is_path_valid(self, path: Path) -> bool:
        """Returns whether a path is valid on the current operating system.

        :param path: the path to validate
        :return: True if the path is valid on the current operating system, False if not
        """
        try:
            # This call fails if the path contains invalid characters
            path.exists()
        except OSError:
            return False

        # On Windows path.exists() doesn't throw for paths like CON/file.txt
        # Trying to create them does raise errors, so we manually validate path components
        if self._platform_manager.is_system_windows():
            # Skip the first component, which contains the drive name
            for component in path.as_posix().split("/")[1:]:
                if component.startswith(" ") or component.endswith(" ") or component.endswith("."):
                    return False

                for reserved_name in ["CON", "PRN", "AUX", "NUL",
                                      "COM1", "COM2", "COM3", "COM4", "COM5", "COM6", "COM7", "COM8", "COM9",
                                      "LPT1", "LPT2", "LPT3", "LPT4", "LPT5", "LPT6", "LPT7", "LPT8", "LPT9"]:
                    if component.upper() == reserved_name or component.upper().startswith(reserved_name + "."):
                        return False

                for forbidden_character in [":", "*", "?", '"', "<", ">", "|"]:
                    if forbidden_character in component:
                        return False

        return True



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import re
from datetime import datetime
from pathlib import Path
from typing import Optional, List, Tuple

import click
import json5
from docker.types import Mount

from lean.click import LeanCommand, PathParameter, ensure_options
from lean.constants import DEFAULT_ENGINE_IMAGE
from lean.container import container
from lean.models.api import QCParameter, QCBacktest
from lean.models.errors import MoreInfoError
from lean.models.optimizer import OptimizationTarget


@click.command(cls=LeanCommand, requires_lean_config=True, requires_docker=True)
@click.argument("project", type=PathParameter(exists=True, file_okay=True, dir_okay=True))
@click.option("--output",
              type=PathParameter(exists=False, file_okay=False, dir_okay=True),
              help="Directory to store results in (defaults to PROJECT/optimizations/TIMESTAMP)")
@click.option("--detach", "-d",
              is_flag=True,
              default=False,
              help="Run the optimization in a detached Docker container and return immediately")
@click.option("--optimizer-config",
              type=PathParameter(exists=True, file_okay=True, dir_okay=False),
              help=f"The optimizer configuration file that should be used")
@click.option("--strategy",
              type=click.Choice(["Grid Search", "Euler Search"], case_sensitive=False),
              help="The optimization strategy to use")
@click.option("--target",
              type=str,
              help="The target statistic of the optimization")
@click.option("--target-direction",
              type=click.Choice(["min", "max"], case_sensitive=False),
              help="Whether the target must be minimized or maximized")
@click.option("--parameter",
              type=(str, float, float, float),
              multiple=True,
              help="The 'parameter min max step' pairs configuring the parameters to optimize")
@click.option("--constraint",
              type=str,
              multiple=True,
              help="The 'statistic operator value' pairs configuring the constraints of the optimization")
@click.option("--release",
              is_flag=True,
              default=False,
              help="Compile C# projects in release configuration instead of debug")
@click.option("--image",
              type=str,
              help=f"The LEAN engine image to use (defaults to {DEFAULT_ENGINE_IMAGE})")
@click.option("--update",
              is_flag=True,
              default=False,
              help="Pull the LEAN engine image before running the optimizer")
def optimize(project: Path,
             output: Optional[Path],
             detach: bool,
             optimizer_config: Optional[Path],
             strategy: Optional[str],
             target: Optional[str],
             target_direction: Optional[str],
             parameter: List[Tuple[str, float, float, float]],
             constraint: List[str],
             release: bool,
             image: Optional[str],
             update: bool) -> None:
    """Optimize a project's parameters locally using Docker.

    \b
    If PROJECT is a directory, the algorithm in the main.py or Main.cs file inside it will be executed.
    If PROJECT is a file, the algorithm in the specified file will be executed.

    By default an interactive wizard is shown letting you configure the optimizer.
    If --optimizer-config or --strategy is given the command runs in non-interactive mode.
    In this mode the CLI does not prompt for input.

    \b
    The --optimizer-config option can be used to specify the configuration to run the optimizer with.
    When using the option it should point to a file like this (the algorithm-* properties should be omitted):
    https://github.com/QuantConnect/Lean/blob/master/Optimizer.Launcher/config.json

    If --strategy is given the optimizer configuration is read from the given options.
    In this case --strategy, --target, --target-direction and --parameter become required.

    \b
    In non-interactive mode the --parameter option can be provided multiple times to configure multiple parameters:
    - --parameter <name> <min value> <max value> <step size>
    - --parameter my-first-parameter 1 10 0.5 --parameter my-second-parameter 20 30 5

    \b
    In non-interactive mode the --constraint option can be provided multiple times to configure multiple constraints:
    - --constraint "<statistic> <operator> <value>"
    - --constraint "Sharpe Ratio >= 0.5" --constraint "Drawdown < 0.25"

    By default the official LEAN engine image is used.
    You can override this using the --image option.
    Alternatively you can set the default engine image for all commands using `lean config set engine-image <image>`.
    """
    project_manager = container.project_manager()
    algorithm_file = project_manager.find_algorithm_file(project)

    if output is None:
        output = algorithm_file.parent / "optimizations" / datetime.now().strftime("%Y-%m-%d_%H-%M-%S")

    optimizer_config_manager = container.optimizer_config_manager()
    config = None

    if optimizer_config is not None and strategy is not None:
        raise RuntimeError("--optimizer-config and --strategy are mutually exclusive")

    if optimizer_config is not None:
        config = json5.loads(optimizer_config.read_text(encoding="utf-8"))

        # Remove keys which are configured in the Lean config
        for key in ["algorithm-type-name", "algorithm-language", "algorithm-location"]:
            config.pop(key, None)
    elif strategy is not None:
        ensure_options(["strategy", "target", "target_direction", "parameter"])

        optimization_strategy = f"QuantConnect.Optimizer.Strategies.{strategy.replace(' ', '')}OptimizationStrategy"
        optimization_target = OptimizationTarget(target=optimizer_config_manager.parse_target(target),
                                                 extremum=target_direction)
        optimization_parameters = optimizer_config_manager.parse_parameters(parameter)
        optimization_constraints = optimizer_config_manager.parse_constraints(constraint)
    else:
        project_config_manager = container.project_config_manager()
        project_config = project_config_manager.get_project_config(algorithm_file.parent)
        project_parameters = [QCParameter(key=k, value=v) for k, v in project_config.get("parameters", {}).items()]

        if len(project_parameters) == 0:
            raise MoreInfoError("The given project has no parameters to optimize",
                                "https://www.lean.io/docs/v2/lean-cli/optimization/parameters")

        optimization_strategy = optimizer_config_manager.configure_strategy(cloud=False)
        optimization_target = optimizer_config_manager.configure_target()
        optimization_parameters = optimizer_config_manager.configure_parameters(project_parameters, cloud=False)
        optimization_constraints = optimizer_config_manager.configure_constraints()

    if config is None:
        # noinspection PyUnboundLocalVariable
        config = {
            "optimization-strategy": optimization_strategy,
            "optimization-strategy-settings": {
                "$type": "QuantConnect.Optimizer.Strategies.StepBaseOptimizationStrategySettings, QuantConnect.Optimizer",
                "default-segment-amount": 10
            },
            "optimization-criterion": {
                "target": optimization_target.target,
                "extremum": optimization_target.extremum.value
            },
            "parameters": [parameter.dict() for parameter in optimization_parameters],
            "constraints": [constraint.dict(by_alias=True) for constraint in optimization_constraints]
        }

    config["optimizer-close-automatically"] = True
    config["results-destination-folder"] = "/Results"

    config_path = output / "optimizer-config.json"
    config_path.parent.mkdir(parents=True, exist_ok=True)
    with config_path.open("w+", encoding="utf-8") as file:
        file.write(json.dumps(config, indent=4) + "\n")

    project_config_manager = container.project_config_manager()
    cli_config_manager = container.cli_config_manager()

    project_config = project_config_manager.get_project_config(algorithm_file.parent)
    engine_image = cli_config_manager.get_engine_image(image or project_config.get("engine-image", None))

    logger = container.logger()
    
    if str(engine_image) != DEFAULT_ENGINE_IMAGE:
        logger.warn(f'A custom engine image: "{engine_image}" is being used!')
        
    lean_config_manager = container.lean_config_manager()
    lean_config = lean_config_manager.get_complete_lean_config("backtesting", algorithm_file, None)

    if not output.exists():
        output.mkdir(parents=True)

    output_config_manager = container.output_config_manager()
    lean_config["algorithm-id"] = str(output_config_manager.get_optimization_id(output))
    lean_config["messaging-handler"] = "QuantConnect.Messaging.Messaging"

    lean_runner = container.lean_runner()
    run_options = lean_runner.get_basic_docker_config(lean_config, algorithm_file, output, None, release, detach)

    run_options["working_dir"] = "/Lean/Optimizer.Launcher/bin/Debug"
    run_options["commands"].append("dotnet QuantConnect.Optimizer.Launcher.dll")
    run_options["mounts"].append(
        Mount(target="/Lean/Optimizer.Launcher/bin/Debug/config.json",
              source=str(config_path),
              type="bind",
              read_only=True)
    )

    container.update_manager().pull_docker_image_if_necessary(engine_image, update)

    project_manager.copy_code(algorithm_file.parent, output / "code")

    success = container.docker_manager().run_image(engine_image, **run_options)

    cli_root_dir = container.lean_config_manager().get_cli_root_directory()
    relative_project_dir = project.relative_to(cli_root_dir)
    relative_output_dir = output.relative_to(cli_root_dir)

    if detach:
        temp_manager = container.temp_manager()
        temp_manager.delete_temporary_directories_when_done = False

        logger.info(
            f"Successfully started optimization for '{relative_project_dir}' in the '{run_options['name']}' container")
        logger.info(f"The output will be stored in '{relative_output_dir}'")
        logger.info("You can use Docker's own commands to manage the detached container")
    elif success:
        optimizer_logs = (output / "log.txt").read_text(encoding="utf-8")
        groups = re.findall(r"ParameterSet: \(([^)]+)\) backtestId '([^']+)'", optimizer_logs)

        if len(groups) > 0:
            optimal_parameters, optimal_id = groups[0]

            optimal_results = json.loads((output / optimal_id / f"{optimal_id}.json").read_text(encoding="utf-8"))
            optimal_backtest = QCBacktest(backtestId=optimal_id,
                                          projectId=1,
                                          status="",
                                          name=optimal_id,
                                          created=datetime.now(),
                                          completed=True,
                                          progress=1.0,
                                          runtimeStatistics=optimal_results["RuntimeStatistics"],
                                          statistics=optimal_results["Statistics"])

            logger.info(f"Optimal parameters: {optimal_parameters.replace(':', ': ').replace(',', ', ')}")
            logger.info(f"Optimal backtest results:")
            logger.info(optimal_backtest.get_statistics_table())

        logger.info(f"Successfully optimized '{relative_project_dir}' and stored the output in '{relative_output_dir}'")
    else:
        raise RuntimeError(
            f"Something went wrong while running the optimization, the output is stored in '{relative_output_dir}'")



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from random import choice


class NameGenerator:
    """The NameGenerator generates random names."""

    def __init__(self) -> None:
        """Creates a new NameGenerator instance."""
        self._verbs = ["Determined", "Pensive", "Adaptable", "Calculating", "Logical", "Energetic", "Creative",
                       "Smooth", "Calm", "Hyper-Active", "Measured", "Fat", "Emotional", "Crying", "Jumping",
                       "Swimming", "Crawling", "Dancing", "Focused", "Well Dressed", "Retrospective", "Hipster",
                       "Square", "Upgraded", "Ugly", "Casual", "Formal", "Geeky", "Virtual", "Muscular", "Alert",
                       "Sleepy"]

        self._colors = ["Red", "Red-Orange", "Orange", "Yellow", "Tan", "Yellow-Green", "Yellow-Green",
                        "Fluorescent Orange", "Apricot", "Green", "Fluorescent Pink", "Sky Blue", "Fluorescent Yellow",
                        "Asparagus", "Blue", "Violet", "Light Brown", "Brown", "Magenta", "Black"]

        self._animals = ["Horse", "Zebra", "Whale", "Tapir", "Barracuda", "Cow", "Cat", "Wolf", "Hamster", "Monkey",
                         "Pelican", "Snake", "Albatross", "Viper", "Guanaco", "Anguilline", "Badger", "Dogfish", "Duck",
                         "Butterfly", "Gaur", "Rat", "Termite", "Eagle", "Dinosaur", "Pig", "Seahorse", "Hornet",
                         "Koala", "Hippopotamus", "Cormorant", "Jackal", "Rhinoceros", "Panda", "Elephant", "Penguin",
                         "Beaver", "Hyena", "Parrot", "Crocodile", "Baboon", "Pony", "Chinchilla", "Fox", "Lion",
                         "Mosquito", "Cobra", "Mule", "Coyote", "Alligator", "Pigeon", "Antelope", "Goat", "Falcon",
                         "Owlet", "Llama", "Gull", "Chicken", "Caterpillar", "Giraffe", "Rabbit", "Flamingo", "Caribou",
                         "Goshawk", "Galago", "Bee", "Jellyfish", "Buffalo", "Salmon", "Bison", "Dolphin", "Jaguar",
                         "Dog", "Armadillo", "Gorilla", "Alpaca", "Kangaroo", "Dragonfly", "Salamander", "Owl", "Bat",
                         "Sheep", "Frog", "Chimpanzee", "Bull", "Scorpion", "Lemur", "Camel", "Leopard", "Fish",
                         "Donkey", "Manatee", "Shark", "Bear", "kitten", "Fly", "Ant", "Sardine"]

    def generate_name(self) -> str:
        """Returns a random name.

        :return: a random name containing multiple words
        """
        return f"{choice(self._verbs)} {choice(self._colors)} {choice(self._animals)}"
